# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: config.py,v 1.2 2004/10/19 01:59:55 nateaune Exp $
"""
from Products.CMFCore.CMFCorePermissions import AddPortalContent
PROJECTNAME = 'Calendaring'
TOOLIDS = ['portal_calendar']
GLOBALS = globals()
ADD_CONTENT_PERMISSION = AddPortalContent

ACTION_ICONS = {'plone':
                {'export_calendar':('subscribe_ical_icon.gif',
                                    'Subscribe to Calendar'),
                 'calendar_day_view':('day_icon.gif',
                                      'Daily'),
                 'calendar_week_view':('week_icon.gif',
                                       'Week'),
                 'calendar_month_view':('month_icon.gif',
                                        'Month'),
                 }
                }



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: test_content.py,v 1.8 2005/01/25 01:03:48 dreamcatcher Exp $
"""
import os, sys
from os import curdir
from os.path import join, abspath, dirname, split
from StringIO import StringIO

try:
    __file__
except NameError:
    # Test was called directly, so no __file__ global exists.
    _prefix = abspath(curdir)
else:
    # Test was called by another test.
    _prefix = abspath(dirname(__file__))

if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))
print _prefix

# Load fixture
import unittest
from doctest import DocTestSuite
from Testing import ZopeTestCase
from Testing.ZopeTestCase import FunctionalDocFileSuite as DocSuite

# Install our product
ZopeTestCase.installProduct('MimetypesRegistry')
ZopeTestCase.installProduct('PortalTransforms')
ZopeTestCase.installProduct('Archetypes')
ZopeTestCase.installProduct('Calendaring')
ZopeTestCase.installProduct('ATContentTypes')

import datetime
from DateTime import DateTime
from Products.CMFPlone.tests import PloneTestCase
from Products.CMFCore.tests.base.utils import has_path
from Products.Calendaring.parser import ICalParseError
from Products.Calendaring.tools import calendar
from Products.Calendaring.Extensions.Install import install_tools
from Products.Calendaring.tests.helpers import sorted
from Products.Calendaring.tests.utils import installType
from Products.Calendaring.common import toTime

class TestImportCalendar(PloneTestCase.PloneTestCase):

    def _setupCalendar(self):
        self.qi = self.portal.portal_quickinstaller
        # Replace tool manually, to test normal events
        install_tools(self.portal, 'Calendaring',
                      [calendar.CalendarTool.meta_type], StringIO())
        self.cal = self.portal.portal_calendar

        # Install CMF Event as Event.
        pt = self.portal.portal_types
        if 'Event' in pt.objectIds():
            pt.manage_delObjects(ids=['Event'])
        pt.manage_addTypeInformation(
            id='Event',
            add_meta_type='Factory-based Type Information',
            typeinfo_name='CMFCalendar: Event (CMF Event)')

        self.cal.setMapping('cmf')

    def afterSetUp(self):
        self.loginPortalOwner()
        self._setupCalendar()
        self.ct = self.portal.portal_catalog
        self.mt = self.portal.portal_membership
        self.wt = self.portal.portal_workflow
        self.mt.createMemberarea()

    def testImport1(self):
        self.file = open(join(_prefix, 'data', 'example.ics'), 'r')
        self.cal.importCalendar(self.file, do_action=True)
        self.file.close()
        result = self.cal.catalog_getevents(year=2002, month=10)
        events = result[28]['eventslist']
        self.assertEquals(len(events), 1)
        self.assertEquals(events[0]['start'], '14:00:00')
        self.assertEquals(events[0]['end'], '15:00:00')
        self.assertEquals(events[0]['title'], 'Coffee with Jason')
        result = self.cal.catalog_getevents(year=2002, month=11)
        events = result[27]['eventslist']
        self.assertEquals(len(events), 1)
        self.assertEquals(events[0]['start'], '12:00:00')
        self.assertEquals(events[0]['end'], '13:00:00')
        self.assertEquals(events[0]['title'], 'Code Review')
        result = self.cal.catalog_getevents(year=2002, month=12)
        events = result[16]['eventslist']
        self.assertEquals(len(events), 1)
        self.assertEquals(events[0]['start'], '20:00:00')
        self.assertEquals(events[0]['end'], '21:00:00')
        self.assertEquals(events[0]['title'], 'Dinner with T')
        gef = self.cal.getEventsForThisDay
        self.assertEquals(len(gef(DateTime(2002, 10, 28))), 1)
        self.assertEquals(len(gef(DateTime(2002, 11, 27))), 1)
        self.assertEquals(len(gef(DateTime(2002, 12, 16))), 1)

    def testImport2(self):
        self.file = open(join(_prefix, 'data', 'tasks.ics'), 'r')
        self.items = self.cal.importCalendar(self.file, do_action=True)
        self.file.close()
        result = self.cal.catalog_getevents(year=2004, month=4)
        events = result[26]['eventslist']
        self.assertEquals(len(events), 1)
        self.assertEquals(events[0]['start'], '15:30:00')
        self.assertEquals(events[0]['end'], '22:30:00')
        self.assertEquals(events[0]['title'], 'Formal Languages Exam')
        item = self.items[0][1]
        self.assertEquals(toTime(item.start_date), '09:00:00')
        self.assertEquals(toTime(item.end_date), '15:00:00')
        self.assertEquals(item.Title(), 'Send Invoice')
        self.assertEquals(self.wt.getInfoFor(item, 'review_state'), 'private')
        gef = self.cal.getEventsForThisDay
        self.assertEquals(len(gef(DateTime(2004, 4, 26))), 1)

    def testImport3(self):
        self.file = open(join(_prefix, 'data', 'frisco.ics'), 'r')
        self.items = self.cal.importCalendar(self.file, do_action=True)
        self.file.close()
        result = self.cal.catalog_getevents(year=2004, month=12)
        events = result[6]['eventslist']
        self.assertEquals(len(events), 1)
        self.assertEquals(events[0]['start'], '20:30:00')
        self.assertEquals(events[0]['end'], '21:30:00')
        self.assertEquals(events[0]['title'], 'Brazilian jazz at Berkeley')
        item = self.items[0][1]
        self.assertEquals(toTime(item.start_date), '20:30:00')
        self.assertEquals(toTime(item.end_date), '21:30:00')
        self.assertEquals(item.Title(), 'Brazilian jazz at Berkeley')
        self.assertEquals(self.wt.getInfoFor(item, 'review_state'), 'published')
        gef = self.cal.getEventsForThisDay
        self.assertEquals(len(gef(DateTime(2004, 12, 6))), 1)

class TestContentTypes(TestImportCalendar):

    def _setupCalendar(self):
        self.qi = self.portal.portal_quickinstaller
        self.qi.installProduct('PortalTransforms')
        self.qi.installProduct('Archetypes')
        self.qi.installProduct('Calendaring')
        self.cal = self.portal.portal_calendar

        # Install ATEvent as Event
        installType(self.portal,
                    'ATEvent',
                    'ATContentTypes',
                    'Event',
                    ignore_actions=True)

        self.cal.setMapping('atct')

    def testImport2(self):
        TestImportCalendar.testImport2(self)
        item = self.items[0][1]
        # not working yet.
        #print item.getLocation()
        #print item.getAtendees()
        #print item.getContact_name()
        #print item.getContact_email()

    def _checkTasks(self, cal, result, items):
        events = result[26]['eventslist']
        self.assertEquals(len(events), 1, events)
        self.assertEquals(events[0]['start'], '15:30:00')
        self.assertEquals(events[0]['end'], '22:30:00')
        self.assertEquals(events[0]['title'], 'Formal Languages Exam')

        item = items[0][1]

        self.failIf(cal.getEvent(item) is None)
        self.assertEquals(toTime(item.start_date), '09:00:00')
        self.assertEquals(toTime(item.end_date), '15:00:00')
        self.assertEquals(item.Title(), 'Send Invoice')
        self.assertEquals(self.wt.getInfoFor(item, 'review_state'), 'private')

        gef = self.cal.getEventsForThisDay
        self.assertEquals(len(gef(DateTime(2004, 4, 26))), 1)
        self.wt.doActionFor(item, 'show')
        self.assertEquals(len(gef(DateTime(2004, 4, 26))), 1)
        self.wt.doActionFor(item, 'submit')
        self.assertEquals(len(gef(DateTime(2004, 4, 26))), 1)
        self.wt.doActionFor(item, 'publish')
        self.assertEquals(len(gef(DateTime(2004, 4, 26))), 2)
        self.wt.doActionFor(item, 'retract')
        self.assertEquals(len(gef(DateTime(2004, 4, 26))), 1)
        self.wt.doActionFor(item, 'hide')
        self.assertEquals(len(gef(DateTime(2004, 4, 26))), 1)

        item = items[1][1]

        self.failIf(cal.getEvent(item) is None)
        self.assertEquals(toTime(item.start_date), '15:30:00')
        self.assertEquals(toTime(item.end_date), '22:30:00')
        self.assertEquals(item.Title(), 'Formal Languages Exam')
        self.assertEquals(self.wt.getInfoFor(item, 'review_state'), 'published')


    def testCalendar(self):
        self.portal.invokeFactory('Calendar', id='test_calendar')
        cal = self.portal.test_calendar
        self.portal.invokeFactory('Calendar', id='test_calendar2')
        cal2 = self.portal.test_calendar2
        self.file = open(join(_prefix, 'data', 'tasks.ics'), 'r')
        self.items = cal.fromFile(self.file)

        self.assertEquals(len(cal.objectIds()), 2)
        self.assertEquals(len(cal2.objectIds()), 0)

        result = self.cal.catalog_getevents(year=2004, month=4)
        self._checkTasks(cal, result, self.items)

        self.file.seek(0)
        cal2.fromFile(self.file)
        self.file.close()

        self.assertEquals(len(cal2.objectIds()), 2)
        cal.update(cal2)
        self.assertEquals(len(cal.objectIds()), 2)
        data = StringIO(cal.asCalendar())
        cal.clear()
        cal2.clear()
        self.assertEquals(len(cal.objectIds()), 0)
        self.items = cal.fromFile(data)
        self.assertEquals(len(cal.objectIds()), 2)
        result = self.cal.catalog_getevents(year=2004, month=4)
        self._checkTasks(cal, result, self.items)

    def testKOrganizer(self):
        self.portal.invokeFactory('Calendar', id='test_calendar')
        cal = self.portal.test_calendar
        self.file = open(join(_prefix, 'data', 'korganizer.ics'), 'r')
        self.items = cal.fromFile(self.file)
        self.file.close()

        item = self.items[0][1]
        value = '"SIDNEI DA SILVA" <sidnei@awkly.org>'
        self.failUnless(value in item.getAttendees())
        self.assertEquals(len(item.getAttendees()), 3, item.attendees)
        self.assertEquals(item.getLocation(), 'Plone.org')

        item = self.items[1][1]
        self.assertEquals(len(item.getAttendees()), 0)
        self.assertEquals(item.getLocation(), 'University')

        item = self.items[2][1]
        self.assertEquals(len(item.getAttendees()), 0)
        self.assertEquals(item.getLocation(), "Mom's House")

        item = self.items[3][1]
        value = '"SIDNEI DA SILVA" <sidnei@awkly.org>'
        self.failUnless(value in item.getAttendees())
        value = 'KITCHEN ASSEMBLER'
        self.failUnless(value in item.getAttendees())
        self.assertEquals(len(item.getAttendees()), 2)
        self.assertEquals(item.getLocation(), 'My Apartment')
        self.assertEquals(item.Title(), 'Assemble Kitchen')
        self.assertEquals(item.Description(),
                          ('Help and supervise the kitchen '
                           'assembling guy. All day :('))

        item = self.items[4][1]
        value = '"SIDNEI DA SILVA" <sidnei@awkly.org>'
        self.failUnless(value in item.getAttendees())
        value = '"PAUL EVERITT" <paul@zope-europe.org>'
        self.failUnless(value in item.getAttendees())
        self.assertEquals(len(item.getAttendees()), 2)
        self.assertEquals(item.getLocation(), 'Europe')
        self.assertEquals(item.Title(), 'Send Invoice')
        self.assertEquals(item.Description(),
                          'Bill for 10h sysadm + 20h development\n')

    def testOutlook1(self):
        self.portal.invokeFactory('Calendar', id='test_calendar')
        cal = self.portal.test_calendar
        self.file = open(join(_prefix, 'data', 'outlook1.ics'), 'r')
        self.items = cal.fromFile(self.file)
        self.file.close()

        item = self.items[0][1]
        self.failIf(item.getAttendees())
        self.assertEquals(item.getLocation(), 'Anywhere')
        self.assertEquals(item.Title(), 'Send event to runyaga')
        self.assertEquals(toTime(item.start_date), '08:00:00')
        self.assertEquals(toTime(item.end_date), '08:30:00')

    def testOutlook2(self):
        self.portal.invokeFactory('Calendar', id='test_calendar')
        cal = self.portal.test_calendar
        self.file = open(join(_prefix, 'data', 'outlook2.ics'), 'r')
        self.items = cal.fromFile(self.file)
        self.file.close()

        item = self.items[0][1]
        self.failIf(item.getAttendees())
        self.assertEquals(item.Title(), 'My Subject')
        self.assertEquals(item.getLocation(), 'My Location')
        self.assertEquals(item.Description(), 'This is a note.\n')
        self.assertEquals(toTime(item.start_date), '00:00:00')
        self.assertEquals(toTime(item.end_date), '00:00:00')
        self.assertEquals(item.start_date, datetime.datetime(2004, 4, 27))
        self.assertEquals(item.end_date, datetime.datetime(2004, 4, 28))
        self.assertEquals(tuple(item.Subject()), ('Webrent2',))

    def testOutlook3(self):
        self.portal.invokeFactory('Calendar', id='test_calendar')
        cal = self.portal.test_calendar
        self.file = open(join(_prefix, 'data', 'outlook3.ics'), 'r')
        self.items = cal.fromFile(self.file)
        self.file.close()

        item = self.items[0][1]
        attendees = item.getAttendees()
        self.failUnless(attendees)
        values = (
            '"Sidnei da Silva" <sidnei@enfoldsystems.com>',
            '"George Alan Runyan" <alan@enfoldsystems.com>',
            '"Toby Roberts" <toby@enfoldsystems.com>',
            '"Andy McKay" <andy@enfoldsystems.com>')
        for value in values:
            self.failUnless(value in attendees, (value, attendees))

        self.assertEquals(item.Title(), 'Test Event: Weekly Meeting')
        self.assertEquals(item.getLocation(), 'Enfold Systems')
        self.assertEquals(item.Description().strip(), '')
        self.assertEquals(toTime(item.start_date), '15:30:00')
        self.assertEquals(toTime(item.end_date), '16:00:00')
        self.assertEquals(item.start_date,
                          datetime.datetime(2005, 10, 21, 15, 30))
        self.assertEquals(item.end_date,
                          datetime.datetime(2005, 10, 21, 16, 0))
        self.assertEquals(tuple(item.Subject()), ())

    def testSunbird(self):
        self.portal.invokeFactory('Calendar', id='test_calendar')
        cal = self.portal.test_calendar
        self.file = open(join(_prefix, 'data', 'sunbird.ics'), 'r')
        self.items = cal.fromFile(self.file)
        self.file.close()

        item = self.items[0][1]
        self.failIf(item.getAttendees())
        self.failIf(item.getLocation())
        self.assertEquals(item.Title(), 'IIS Caching')
        self.assertEquals(toTime(item.start_date), '16:00:00')
        self.assertEquals(toTime(item.end_date), '16:00:00')


class TestExport(PloneTestCase.FunctionalTestCase):

    def afterSetUp(self):
        self.loginPortalOwner()
        self.qi = self.portal.portal_quickinstaller
        self.qi.installProduct('PortalTransforms')
        self.qi.installProduct('Archetypes')
        self.qi.installProduct('Calendaring')
        self.cal = self.portal.portal_calendar

        # Install ATEvent as Event
        installType(self.portal,
                    'ATEvent',
                    'ATContentTypes',
                    'Event',
                    ignore_actions=True)

        self.cal.setMapping('atct')


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestContentTypes))
    suite.addTest(unittest.makeSuite(TestImportCalendar))
    functional = [
        (TestExport, 'export.txt')
        ]
    for test, f in functional:
        suite.addTest(DocSuite(f, test_class=test,
                               package='Products.Calendaring.tests',
                               globs=globals()))
    return suite

if __name__ == '__main__':
    framework(descriptions=1, verbosity=1)



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: _base.py,v 1.5 2005/01/11 00:26:34 dreamcatcher Exp $
"""
import urllib
from datetime import datetime
from Acquisition import Implicit, aq_inner, aq_parent
from ComputedAttribute import ComputedAttribute
from Products.CMFCore.utils import getToolByName
from Products.Calendaring.model import makeRange, DAY, getView, SECOND
from Products.Calendaring.common import dt2DT, DT2dt, toSeconds

ARG_NAMES = ['cyear', 'cmonth', 'cday', 'hour', 'minute', 'second']

class ModelView(Implicit):
    """ Base ModelView class for support of calendar rendering """

    __allow_access_to_unprotected_subobjects__ = 1

    unit = 'day'
    _sub = 'day_view'
    name = 'day_view'
    _args = 2
    _leap = DAY
    _skip = True

    # DWM: Added line to init to check the portal calendar's calendar
    # types.  Would be preferable to check the calendar container for
    # registered event types
    #calendar_types = ['Event']

    def __init__(self, dt=None, start=None, REQUEST=None):
        if dt is None:
            dt = datetime.today()
        if start is None:
            start = dt.replace(hour=0, minute=0, second=0, microsecond=0)
            if REQUEST is not None:
                start = dt.replace(hour=0, minute=0, second=0, microsecond=0)
                d = {}
                for k in ('cyear', 'cmonth', 'cday'):
                    if REQUEST.has_key(k):
                        d[k[1:]] = int(REQUEST.get(k))
                if d:
                    start = start.replace(**d)
        self.today = dt
        # So we always have a datetime inside
        # the period that we are looking at.
        # XXX DC: This is not guaranteed!
        self.cur = start
        self.start, self.end, self.resolution = makeRange(start, self.unit)

        # set calendar_types to display from portal_calendar


    def _context(self):
        return aq_parent(self)

    context = ComputedAttribute(_context, 1)

    def _factory_hook(self):
        return getView(self._sub)

    def _sub_factory(self):
        return self._factory_hook()

    sub_factory = ComputedAttribute(_sub_factory)

    def _parts(self):
        factory = self.sub_factory
        one = factory(self.today, self.start)
        end = one.end
        yield one.__of__(self.context)
        while end < self.end:
            sub = factory(self.today, end)
            end = sub.end
            yield sub.__of__(self.context)

    parts = ComputedAttribute(_parts, 1)

    def next(self):
        factory = self.__class__
        view = factory(self.today, self.end + self._leap)
        return view.__of__(self.context)

    def previous(self):
        factory = self.__class__
        leap = self._leap
        if self._skip:
            leap += DAY
        view = factory(self.today, self.start - leap)
        return view.__of__(self.context)

    def _make_url(self, date):
        request = self.REQUEST
        prefix = ''
        if hasattr(self, 'aq_parent'):
            prefix = self.context.absolute_url()
        args = dict(zip(ARG_NAMES[:self._args],
                        date.timetuple()[:self._args]))
        parts = filter(None, [prefix, self.name])
        return '%s?%s' % ('/'.join(parts), urllib.urlencode(args))

    def next_url(self):
        return self._make_url(self.next().start)

    def previous_url(self):
        return self._make_url(self.previous().start)

    def _day(self):
        return self.today.day

    day = ComputedAttribute(_day)

    def _month(self):
        return self.today.month

    month = ComputedAttribute(_month)

    def _year(self):
        return self.today.year

    year = ComputedAttribute(_year)

    def _current(self):
        return self.start <= self.today <= self.end

    current = ComputedAttribute(_current)

    def __str__(self):
        name = self.__class__.__name__
        args = ('name', self.name,
                'unit', self.unit,
                'sub', self._sub,
                'today', self.today,
                'start', self.start,
                'end', self.end,
                'resolution', self.resolution)
        return '%s%r' % (name, args)

    __repr__ = __str__

    def index_html(self):
        """ Get template from skins tool """
        st = getToolByName(self, 'portal_skins')
        template = st.restrictedTraverse(self.name)
        template = template.aq_base.__of__(self.context)
        return template(view=self)

    def getDateTimes(self):
        """
        Return the zope DateTime objects for self.today self.start
        and self.end
        """
        return {'today': dt2DT(self.today),
                'current': dt2DT(self.cur),
                'start': dt2DT(self.start),
                'end': dt2DT(self.end - SECOND),
                }

    def _ongoingEvents(self):
        pc_tool = getToolByName(self, 'portal_calendar')
        calendar_types = pc_tool.calendar_types

        ct = getToolByName(self, 'portal_catalog')
        start = dt2DT(self.start)
        end = dt2DT(self.start + SECOND)
        res = ct(portal_type=calendar_types,
                  start={'query': start,
                         'range': 'max'},
                  end={'query': end,
                       'range': 'min'},
                 sort_on='start')
        return tuple(res)

    ongoing_events = ComputedAttribute(_ongoingEvents, 1)

    def _events(self):
        ct = getToolByName(self, 'portal_catalog')
        pc_tool = getToolByName(self, 'portal_calendar')
        calendar_types = pc_tool.calendar_types
        res = ct(portal_type=calendar_types,
                  start={'query': [dt2DT(self.start),dt2DT(self.end - SECOND)],
                         'range': 'min:max'},
                 sort_on='start')
        return tuple(res)

    events = ComputedAttribute(_events, 1)

    def computeDuration(self, event):
        """
        Given an event computes the duration as an integer multiple of our
        sub_view ie if we're on the day vew returns the number of hours the
        event will last
        """
        start = DT2dt(event.start)
        end = DT2dt(event.end)
        delta = end - start
        # print  '----------->', toSeconds(self.resolution), self.resolution
        # _robert_ prevent division by 0
        return round(toSeconds(delta) / (toSeconds(self.resolution) or 1))



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: __init__.py,v 1.2 2004/09/16 13:46:09 dreamcatcher Exp $
"""
from _cal import makeRange, DAY, WEEKDAY, WEEK, MONTH, YEAR, HOUR, SECOND

_views = {}

def registerView(component, name=None):
    if name is None:
        name = component.name
    _views[name] = component

def getView(name):
    return _views[name]

def queryView(name, default=None):
    return _views.get(name, default)

from _base import ModelView



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Some portions of this module are Copyright Shuttleworth Foundation.
# The original copyright statement is reproduced below.
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2003 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Helper functions for unit tests.

$Id: helpers.py,v 1.1.1.1 2004/06/19 14:53:55 dreamcatcher Exp $
"""

import difflib
import cgi
import libxml2

from Products.Calendaring.common import dedent  # it used to live here


def unidiff(old, new, oldlabel="expected output", newlabel="actual output"):
    """Display a compact unified diff between old text and new text.

    Bugs: does not work when old or new is an empty string (this seems to
    be a bug/limitation of the Python 2.3 difflib module).
    """
    return "\n".join(difflib.unified_diff(old.splitlines(), new.splitlines(),
                                          oldlabel, newlabel, lineterm=''))


def diff(old, new, oldlabel="expected output", newlabel="actual output"):
    """Display a unified diff between old text and new text."""
    old = old.splitlines()
    new = new.splitlines()

    diff = ['--- %s' % oldlabel, '+++ %s' % newlabel]

    def dump(tag, x, lo, hi):
        for i in xrange(lo, hi):
            diff.append(tag + x[i])

    differ = difflib.SequenceMatcher(a=old, b=new)
    for tag, alo, ahi, blo, bhi in differ.get_opcodes():
        if tag == 'replace':
            dump('-', old, alo, ahi)
            dump('+', new, blo, bhi)
        elif tag == 'delete':
            dump('-', old, alo, ahi)
        elif tag == 'insert':
            dump('+', new, blo, bhi)
        elif tag == 'equal':
            dump(' ', old, alo, ahi)
        else:
            raise AssertionError('unknown tag %r' % tag)
    return "\n".join(diff)


def sorted(l):
    l = list(l) # make a copy
    l.sort()
    return l


def normalize_xml(xml, recursively_sort=()):
    """Normalizes an XML document.

    The idea is that two semantically equivalent XML documents should be
    normalized into the same canonical representation.  Therefore if two
    documents compare equal after normalization, they are semantically
    equivalent.

    The canonical representation used here has nothing to do with W3C Canonical
    XML.

    This function normalizes indentation, whitespace and newlines (except
    inside text nodes), element attribute order, expands character references,
    expands shorthand notation of empty XML elements ("<br/> becomes
    "<br></br>").

    If an element has an attribute test:sort="children", the attribute is
    removed and its immediate child nodes are sorted textually.  If the
    attribute value is test:sort="recursively", the sorting happens at
    all levels (unless specifically prohibited with test:sort="not").

    If recursively_sort is given, it is a sequence of tags that will have
    test:sort="recursively" automatically appended to their attribute lists in
    the text.  Use it when you cannot or do not want to modify the XML document
    itself.

    Caveats:
     - normalize_xml does not deal well with text nodes
     - normalize_xml does not help when different prefixes are used for the
       same namespace
     - normalize_xml does not handle all XML features (CDATA sections, inline
       DTDs, processing instructions, comments)
    """

    class Document:

        def __init__(self):
            self.children = []
            self.sort_recursively = False

        def render(self, level=0):
            result = []
            for child in self.children:
                result.append(child.render(level))
            return ''.join(result)

    class Element:

        def __init__(self, parent, tag, attrlist, sort=False,
                     sort_recursively=False):
            self.parent = parent
            self.tag = tag
            self.attrlist = attrlist
            self.children = []
            self.sort = sort
            self.sort_recursively = sort_recursively

        def render(self, level):
            result = []
            indent = '  ' * level
            line = '%s<%s' % (indent, self.tag)
            for attr in self.attrlist:
                if len(line + attr) < 78:
                    line += attr
                else:
                    result.append(line)
                    result.append('\n')
                    line = '%s %s%s' % (indent, ' ' * len(self.tag), attr)
            if self.children:
                s = ''.join([child.render(level+1) for child in self.children])
            else:
                s = ''
            if s:
                result.append('%s>\n' % line)
                result.append(s)
                result.append('%s</%s>\n' % (indent, self.tag))
            else:
                result.append('%s/>\n' % line)
            return ''.join(result)

        def finalize(self):
            if self.sort:
                self.children.sort(lambda x, y: cmp(x.key, y.key))
            self.key = self.render(0)

    class Text:

        def __init__(self, data):
            self.data = data
            self.key = None

        def render(self, level):
            data = cgi.escape(self.data.strip())
            if data:
                indent = '  ' * level
                return ''.join(['%s%s\n' % (indent, line.strip())
                                for line in data.splitlines()])
            else:
                return ''

    class Handler:

        def __init__(self):
            self.level = 0
            self.result = []
            self.root = self.cur = Document()
            self.last_text = None

        def startElement(self, tag, attrs):
            sort = sort_recursively = self.cur.sort_recursively
            if attrs:
                if 'test:sort' in attrs:
                    value = attrs['test:sort']
                    del attrs['test:sort']
                    if value == 'children':
                        sort = True
                    elif value == 'recursively':
                        sort = sort_recursively = True
                    elif value == 'not':
                        sort = sort_recursively = False
                attrlist = attrs.items()
                attrlist.sort()
                attrlist = [' %s="%s"' % (k, cgi.escape(v, True))
                            for k, v in attrlist]
            else:
                attrlist = []
            child = Element(self.cur, tag, attrlist, sort=sort,
                            sort_recursively=sort_recursively)
            self.cur.children.append(child)
            self.cur = child
            self.last_text = None

        def endElement(self, tag):
            self.cur.finalize()
            self.cur = self.cur.parent
            self.last_text = None

        def characters(self, data):
            if self.last_text is not None:
                self.last_text.data += data
            else:
                self.last_text = Text(data)
                self.cur.children.append(self.last_text)

        def render(self):
            return self.root.render()

    for tag in recursively_sort:
        xml = xml.replace('<%s' % tag,
                          '<%s test:sort="recursively"' % tag)
    try:
        handler = Handler()
        ctx = libxml2.createPushParser(handler, "", 0, "")
        ret = ctx.parseChunk(xml, len(xml), True)
        if ret:
            return "PARSE ERROR: %r\n%s" % (ret, xml)
        return ''.join(handler.render())
    except libxml2.parserError, e:
        return "ERROR: %s" % e



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Some portions of this module are Copyright Shuttleworth Foundation.
# The original copyright statement is reproduced below.
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2003 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: test_calevent.py,v 1.2 2004/09/16 13:46:09 dreamcatcher Exp $
"""

import os, sys

if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

import unittest
from doctest import DocTestSuite
from Testing import ZopeTestCase

from datetime import date, timedelta, datetime
from Products.Calendaring.tests.utils import EqualsSortedMixin
from Products.Calendaring.tests.helpers import sorted
from Products.Calendaring.parser import Calendar
from Products.Calendaring.parser import CalendarEvent
from Products.Calendaring.parser import ICalReader, ICalParseError

class TestCalendar(unittest.TestCase, EqualsSortedMixin):

    def test_iter(self):
        cal = Calendar()
        self.assertEqual(list(cal), [])

        ev1 = CalendarEvent(datetime(2003, 11, 25, 10, 0),
                            timedelta(minutes=10),
                            "English")

        cal.addEvent(ev1)
        self.assertEqual(list(cal), [ev1])

    def makeCal(self, events):
        cal = Calendar()
        for event in events:
            cal.addEvent(event)
        return cal

    def test_byDate(self):

        ev1 = CalendarEvent(datetime(2003, 11, 25, 10, 0),
                            timedelta(minutes=10),
                            "English")
        ev2 = CalendarEvent(datetime(2003, 11, 25, 12, 0),
                            timedelta(minutes=10),
                            "Latin")
        ev3 = CalendarEvent(datetime(2003, 11, 26, 10, 0),
                            timedelta(minutes=10),
                            "German")

        cal = self.makeCal([ev1, ev2, ev3])
        self.assertEqual(list(cal.byDate(date(2003, 11, 26))), [ev3])

        # event end date within the period
        ev4 = CalendarEvent(datetime(2003, 11, 25, 10, 0),
                            timedelta(1),
                            "Solar eclipse")
        cal = self.makeCal([ev1, ev2, ev3, ev4])
        self.assertEqualSorted(list(cal.byDate(date(2003, 11, 26))),
                               [ev3, ev4])

        # calendar daterange is within the event period
        ev4 = CalendarEvent(datetime(2003, 11, 25, 10, 0),
                            timedelta(2),
                            "Solar eclipse")
        cal = self.makeCal([ev1, ev2, ev3, ev4])
        self.assertEqualSorted(list(cal.byDate(date(2003, 11, 26))),
                               [ev3, ev4])

        # only the event start date falls within the period
        ev4 = CalendarEvent(datetime(2003, 11, 26, 10, 0),
                            timedelta(2),
                            "Solar eclipse")
        cal = self.makeCal([ev1, ev2, ev3, ev4])
        self.assertEqualSorted(list(cal.byDate(date(2003, 11, 26))),
                               [ev3, ev4])

        # the event is after the period
        ev4 = CalendarEvent(datetime(2003, 11, 27, 10, 0),
                            timedelta(2),
                            "Solar eclipse")
        cal = self.makeCal([ev1, ev2, ev3, ev4])
        self.assertEqualSorted(list(cal.byDate(date(2003, 11, 26))),
                               [ev3])

    def test_clear(self):
        ev1 = CalendarEvent(datetime(2003, 11, 25, 10, 0),
                            timedelta(minutes=10),
                            "English")
        ev2 = CalendarEvent(datetime(2003, 11, 25, 12, 0),
                            timedelta(minutes=10),
                            "Latin")
        cal = self.makeCal([ev1, ev2])
        cal.clear()
        self.assertEquals(list(cal), [])

    def test_removeEvent(self):
        ev1 = CalendarEvent(datetime(2003, 11, 25, 10, 0),
                            timedelta(minutes=10),
                            "English")
        ev2 = CalendarEvent(datetime(2003, 11, 25, 12, 0),
                            timedelta(minutes=10),
                            "Latin")
        cal = self.makeCal([ev1, ev2])
        cal.removeEvent(CalendarEvent(datetime(2003, 11, 25, 10, 0),
                                      timedelta(minutes=10),
                                      "English"))
        self.assertEquals(list(cal), [ev2])

    def test_update(self):
        ev1 = CalendarEvent(datetime(2003, 11, 25, 10, 0),
                            timedelta(minutes=10),
                            "English")
        ev2 = CalendarEvent(datetime(2003, 11, 25, 12, 0),
                            timedelta(minutes=10),
                            "Latin")
        ev3 = CalendarEvent(datetime(2003, 11, 26, 10, 0),
                            timedelta(minutes=10),
                            "German")
        cal = self.makeCal([ev1, ev2])
        cal.update(self.makeCal([ev2, ev3]))
        self.assertEquals(sorted(list(cal)), sorted([ev1, ev2, ev3]))



class TestCalendarEvent(unittest.TestCase):

    def test(self):
        owner = object()
        context = object()

        ce = CalendarEvent(datetime(2003, 11, 25, 12, 0),
                           timedelta(minutes=10),
                           "reality check")

        ce1 = CalendarEvent(datetime(2003, 11, 25, 12, 0),
                            timedelta(minutes=10),
                            "reality check")
        ce2 = CalendarEvent(datetime(2003, 11, 25, 12, 0),
                            timedelta(minutes=10),
                            "realty check")
        ce3 = CalendarEvent(datetime(2003, 11, 25, 12, 1),
                            timedelta(minutes=10),
                            "reality check")
        ce4 = CalendarEvent(datetime(2003, 11, 25, 12, 0),
                            timedelta(minutes=11),
                            "reality check")
        ce5 = CalendarEvent(datetime(2003, 11, 25, 12, 0),
                            timedelta(minutes=10),
                            "reality check", owner=owner)
        ce6 = CalendarEvent(datetime(2003, 11, 25, 12, 0),
                            timedelta(minutes=10),
                            "reality check", owner=owner,
                            context=context)
        self.assertEquals(ce, ce1)
        self.assertNotEquals(ce, ce2)
        self.assertNotEquals(ce, ce3)
        self.assertNotEquals(ce, ce4)
        self.assertNotEquals(ce, ce5)
        self.assertNotEquals(ce5, ce6)
        # self.assertRaises(AttributeError, setattr, ce, 'dtstart', 'not-ro')
        # self.assertRaises(AttributeError, setattr, ce, 'duration', 'not-ro')
        # self.assertRaises(AttributeError, setattr, ce, 'title', 'not-ro')

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestCalendar))
    suite.addTest(unittest.makeSuite(TestCalendarEvent))
    return suite

if __name__ == '__main__':
    framework(descriptions=1, verbosity=1)



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: __init__.py,v 1.3 2004/09/17 13:58:25 dreamcatcher Exp $
"""

# Kickstart Extensions.
import Products.Calendaring.Extensions.Install

from Products.Archetypes.public import *
from Products.Calendaring.config import *
from Products.CMFCore import utils as cmf_utils
from Products.CMFCore.DirectoryView import registerDirectory
from Products.CMFPlone import PloneUtilities as plone_utils

registerDirectory('skins', GLOBALS)

def initialize(context):

    from Products.Calendaring.tools import calendar
    from Products.Calendaring.model import year, month, day, hour, week

    tools = (calendar.CalendarTool,)

    plone_utils.ToolInit('Calendaring Tools', tools=tools,
                         product_name='Calendaring', icon='tool.gif',
                         ).initialize(context)
    del calendar

    from Products.Calendaring.content import calendar

    # register archetypes content with the machinery
    content_types, constructors, ftis = process_types(listTypes(PROJECTNAME),
                                                      PROJECTNAME)
    cmf_utils.ContentInit(
        PROJECTNAME + ' Content',
        content_types = content_types,
        permission = ADD_CONTENT_PERMISSION,
        extra_constructors = constructors,
        fti = ftis).initialize(context)



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Some portions of this module are Copyright Shuttleworth Foundation.
# The original copyright statement is reproduced below.
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2003 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Things common to the SchoolTool server and clients.

$Id: common.py,v 1.2 2004/09/17 13:58:25 dreamcatcher Exp $
"""

import re
import locale
import datetime
from DateTime import DateTime

__metaclass__ = type

def dt2DT(date):
    return DateTime(*date.timetuple()[:6])

def DT2dt(date):
    # seconds (parts[6]) is a float, so we map to int
    args = map(int, date.parts()[:6])
    return datetime.datetime(*args)

def toTime(date):
    if isinstance(date, datetime.datetime):
        date = dt2DT(date)
    return date.Time()

def toSeconds(td):
    """ Converts a timedelta to an integer
    representing the number of minutes
    """
    return td.seconds + td.days * 86400

def parse_date(value):
    """Parse a ISO-8601 YYYY-MM-DD date value.

    Examples:

        >>> parse_date('2003-09-01')
        datetime.date(2003, 9, 1)
        >>> parse_date('20030901')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '20030901'
        >>> parse_date('2003-IX-01')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '2003-IX-01'
        >>> parse_date('2003-09-31')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '2003-09-31'
        >>> parse_date('2003-09-30-15-42')
        Traceback (most recent call last):
          ...
        ValueError: Invalid date: '2003-09-30-15-42'

    """
    try:
        y, m, d = map(int, value.split('-'))
        return datetime.date(y, m, d)
    except ValueError:
        raise ValueError("Invalid date: %r" % value)


def parse_time(value):
    """Parse a ISO 8601 HH:MM time value.

    Examples:

        >>> parse_time('01:25')
        datetime.time(1, 25)
        >>> parse_time('9:15')
        datetime.time(9, 15)
        >>> parse_time('12:1')
        datetime.time(12, 1)
        >>> parse_time('00:00')
        datetime.time(0, 0)
        >>> parse_time('23:59')
        datetime.time(23, 59)
        >>> parse_time('24:00')
        Traceback (most recent call last):
          ...
        ValueError: Invalid time: '24:00'
        >>> parse_time('06:30PM')
        Traceback (most recent call last):
          ...
        ValueError: Invalid time: '06:30PM'

    """
    try:
        h, m = map(int, value.split(':'))
        return datetime.time(h, m)
    except ValueError:
        raise ValueError("Invalid time: %r" % value)


def parse_datetime(s):
    """Parse a ISO 8601 date/time value.

    Only a small subset of ISO 8601 is accepted:

      YYYY-MM-DD HH:MM:SS
      YYYY-MM-DD HH:MM:SS.ssssss
      YYYY-MM-DDTHH:MM:SS
      YYYY-MM-DDTHH:MM:SS.ssssss

    Returns a datetime.datetime object without a time zone.

    Examples:

        >>> parse_datetime('2003-04-05 11:22:33.456789')
        datetime.datetime(2003, 4, 5, 11, 22, 33, 456789)

        >>> parse_datetime('2003-04-05 11:22:33.456')
        datetime.datetime(2003, 4, 5, 11, 22, 33, 456000)

        >>> parse_datetime('2003-04-05 11:22:33.45678999')
        datetime.datetime(2003, 4, 5, 11, 22, 33, 456789)

        >>> parse_datetime('01/02/03')
        Traceback (most recent call last):
          ...
        ValueError: Bad datetime: 01/02/03

    """
    m = re.match("(\d+)-(\d+)-(\d+)[ T](\d+):(\d+):(\d+)([.](\d+))?$", s)
    if not m:
        raise ValueError("Bad datetime: %s" % s)
    ssssss = m.groups()[7]
    if ssssss:
        ssssss = int((ssssss + "00000")[:6])
    else:
        ssssss = 0
    y, m, d, hh, mm, ss = map(int, m.groups()[:6])
    return datetime.datetime(y, m, d, hh, mm, ss, ssssss)


def dedent(text):
    r"""Remove leading indentation from triple-quoted strings.

    Example:

        >>> dedent('''
        ...     some text
        ...     is here
        ...        with maybe some indents
        ...     ''')
        ...
        'some text\nis here\n   with maybe some indents\n'

    Corner cases (mixing tabs and spaces, lines that are indented less than
    the first line) are not handled yet.
    """
    lines = text.splitlines()
    first, limit = 0, len(lines)
    while first < limit and not lines[first]:
        first += 1
    if first >= limit:
        return ''
    firstline = lines[first]
    indent, limit = 0, len(firstline)
    while indent < limit and firstline[indent] in (' ', '\t'):
        indent += 1
    return '\n'.join([line[indent:] for line in lines[first:]])


def to_unicode(s):
    r"""Convert an UTF-8 string to Unicode.

    Example:

        >>> to_unicode('\xc4\x84\xc5\xbeuol\xc5\xb3')
        u'\u0104\u017euol\u0173'

    For convenience, to_unicode accepts None as the argument value.  This makes
    it easier to use it with libxml2 functions such as nsProp, which return
    None for missing attribute values.

        >>> to_unicode(None) is None
        True

    """
    if s is None:
        return None
    else:
        return unicode(s, 'UTF-8')


locale_charset = locale.getpreferredencoding()


def to_locale(us):
    r"""Convert a Unicode string to the current locale encoding.

    Example:

        >>> from schooltool import common
        >>> old_locale_charset = common.locale_charset

        >>> common.locale_charset = 'UTF-8'
        >>> to_locale(u'\u263B')
        '\xe2\x98\xbb'

        >>> common.locale_charset = 'ASCII'
        >>> to_locale(u'Unrepresentable: \u263B')
        'Unrepresentable: ?'

        >>> locale_charset = old_locale_charset

    """
    return us.encode(locale_charset, 'replace')


def from_locale(s):
    r"""Convert an 8-bit string in locale encoding to Unicode.

    Example:

        >>> from schooltool import common
        >>> old_locale_charset = common.locale_charset

        >>> from_locale('xyzzy')
        u'xyzzy'

        >>> common.locale_charset = 'UTF-8'
        >>> from_locale('\xe2\x98\xbb')
        u'\u263b'

        >>> locale_charset = old_locale_charset

    """
    return unicode(s, locale_charset)


class StreamWrapper:
    r"""Unicode-friendly wrapper for writable file-like objects.

    Here the terms 'encoding' and 'charset' are used interchangeably.

    The main use case for StreamWrapper is wrapping sys.stdout and sys.stder so
    that you can forget worrying about charsets of your data.

        >>> from StringIO import StringIO
        >>> from schooltool import common
        >>> old_locale_charset = common.locale_charset
        >>> common.locale_charset = 'UTF-8'

        >>> sw = StreamWrapper(StringIO())
        >>> print >> sw, u"Hello, world! \u00b7\u263b\u00b7"
        >>> sw.stm.getvalue()
        'Hello, world! \xc2\xb7\xe2\x98\xbb\xc2\xb7\n'

    By default printing Unicode strings to stdout/stderr will raise Unicode
    errors if the stream encoding does not include some characters you are
    printing.  StreamWrapper will replace unconvertable characters to question
    marks, therefore you should only use it for informative messages where such
    loss of information is acceptable.

        >>> common.locale_charset = 'US-ASCII'
        >>> sw = StreamWrapper(StringIO())
        >>> print >> sw, u"Hello, world! \u00b7\u263b\u00b7"
        >>> sw.stm.getvalue()
        'Hello, world! ???\n'

    StreamWrapper converts all unicode strings that are written to it to the
    encoding defined in the wrapped stream's 'encoding' attribute, or, if that
    is None, to the locale encoding.  Typically the stream's encoding attribute
    is set when the stream is connected to a console device, and None when the
    stream is connected to a file.  On Unix systems the console encoding
    matches the locale charset, but on Win32 systems they differ.

        >>> s = StringIO()
        >>> s.encoding = 'ISO-8859-1'
        >>> sw = StreamWrapper(s)
        >>> print >> sw, u"Hello, world! \u00b7\u263b\u00b7"
        >>> sw.stm.getvalue()
        'Hello, world! \xb7?\xb7\n'

    You can print other kinds of objects:

        >>> sw = StreamWrapper(StringIO())
        >>> print >> sw, 1, 2,
        >>> print >> sw, 3
        >>> sw.stm.getvalue()
        '1 2 3\n'

    but not 8-bit strings:

        >>> sw = StreamWrapper(StringIO())
        >>> print >> sw, "\xff"
        Traceback (most recent call last):
          ...
        UnicodeDecodeError: 'ascii' codec can't decode byte 0xff in position 0: ordinal not in range(128)

    In addition to 'write' StreamWrapper provides 'flush' and 'writelines'

        >>> sw = StreamWrapper(StringIO())
        >>> sw.write('xyzzy\n')
        >>> sw.flush()
        >>> sw.writelines(['a', 'b', 'c', 'd'])
        >>> sw.stm.getvalue()
        'xyzzy\nabcd'

    """

    def __init__(self, stm):
        self.stm = stm
        self.encoding = getattr(stm, 'encoding', None)
        if self.encoding is None:
            self.encoding = locale_charset

    def write(self, obj):
        self.stm.write(obj.encode(self.encoding, 'replace'))

    def flush(self):
        self.stm.flush()

    def writelines(self, seq):
        for obj in seq:
            self.write(obj)


class UnicodeAwareException(Exception):
    r"""Unicode-friendly exception.

        >>> e1 = UnicodeAwareException(u"\u2639")
        >>> unicode(e1)
        u'\u2639'

        >>> e2 = UnicodeAwareException(u"\u2639", e1)
        >>> unicode(e2)
        u'\u2639 \u2639'

    """

    def __unicode__(self):
        return u" ".join(map(unicode, self.args))



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Some portions of this module are Copyright Shuttleworth Foundation.
# The original copyright statement is reproduced below.
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2003 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
SchoolTool calendaring stuff.

$Id: parser.py,v 1.4 2005/01/25 01:03:38 dreamcatcher Exp $
"""

import re
import datetime
from sets import Set
Persistent = 0
# make parser available outside of zope
try:
    from Persistence import Persistent
except:
    class Persistent:
        pass

class DateRange:

    def __init__(self, first, last):
        self.first = first
        self.last = last
        if last < first:
            # import timemachine
            raise ValueError("Last date %r less than first date %r" %
                             (last, first))

    def __iter__(self):
        date = self.first
        while date <= self.last:
            yield date
            date += datetime.date.resolution

    def __len__(self):
        return (self.last - self.first).days + 1

    def __contains__(self, date):
        return self.first <= date <= self.last


#
# iCalendar parsing
#


class ICalParseError(Exception):
    """Invalid syntax in an iCalendar file."""


def parse_text(value):
    r"""Parse iCalendar TEXT value.

    >>> parse_text('Foo')
    'Foo'
    >>> parse_text('\\\\')
    '\\'
    >>> parse_text('\\;')
    ';'
    >>> parse_text('\\,')
    ','
    >>> parse_text('\\n')
    '\n'
    >>> parse_text('A string with\\; some\\\\ characters\\nin\\Nit')
    'A string with; some\\ characters\nin\nit'
    >>> parse_text('Unterminated \\')
    Traceback (most recent call last):
      ...
    IndexError: string index out of range
    """
    if '\\' not in value:
        return value
    out = []
    prev = 0
    while True:
        idx = value.find('\\', prev)
        if idx == -1:
            break
        out.append(value[prev:idx])
        if value[idx + 1] in 'nN':
            out.append('\n')
        else:
            out.append(value[idx + 1])
        prev = idx + 2
    out.append(value[prev:])
    return "".join(out)


def ical_text(value):
    r"""Format value according to iCalendar TEXT escaping rules.

    >>> ical_text('Foo')
    'Foo'
    >>> ical_text('\\')
    '\\\\'
    >>> ical_text(';')
    '\\;'
    >>> ical_text(',')
    '\\,'
    >>> ical_text('\n')
    '\\n'
    >>> ical_text('\r\n')
    '\\n'
    """
    return (value.replace('\\', '\\\\').replace(';', '\\;').replace(',', '\\,')
                 .replace('\r\n', '\n').replace('\n', '\\n'))


def parse_duration(value):
    """Parse iCalendar DURATION value.  Returns a timedelta instance.

    >>> parse_duration('+P11D')
    datetime.timedelta(11)
    >>> parse_duration('-P2W')
    datetime.timedelta(-14)
    >>> parse_duration('P1DT2H3M4S')
    datetime.timedelta(1, 7384)
    >>> parse_duration('P1DT2H3M')
    datetime.timedelta(1, 7380)
    >>> parse_duration('P1DT2H')
    datetime.timedelta(1, 7200)
    >>> parse_duration('PT2H')
    datetime.timedelta(0, 7200)
    >>> parse_duration('PT2H3M4S')
    datetime.timedelta(0, 7384)
    >>> parse_duration('PT3M4S')
    datetime.timedelta(0, 184)
    >>> parse_duration('PT22S')
    datetime.timedelta(0, 22)
    >>> parse_duration('')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar duration: ''
    >>> parse_duration('xyzzy')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar duration: 'xyzzy'
    >>> parse_duration('P')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar duration: 'P'
    >>> parse_duration('P1WT2H')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar duration: 'P1WT2H'
    """
    date_part = r'(\d+)D'
    time_part = r'T(?:(\d+)H)?(?:(\d+)M)?(?:(\d+)S)?'
    datetime_part = '(?:%s)?(?:%s)?' % (date_part, time_part)
    weeks_part = r'(\d+)W'
    duration_rx = re.compile(r'([-+]?)P(?:%s|%s)$'
                             % (weeks_part, datetime_part))
    match = duration_rx.match(value)
    if match is None:
        raise ValueError('Invalid iCalendar duration: %r' % value)
    sign, weeks, days, hours, minutes, seconds = match.groups()
    if weeks:
        value = datetime.timedelta(weeks=int(weeks))
    else:
        if (days is None and hours is None
            and minutes is None and seconds is None):
            raise ValueError('Invalid iCalendar duration: %r'
                             % value)
        value = datetime.timedelta(days=int(days or 0),
                                   hours=int(hours or 0),
                                   minutes=int(minutes or 0),
                                   seconds=int(seconds or 0))
    if sign == '-':
        value = -value
    return value


def ical_duration(value):
    """Format a timedelta as an iCalendar DURATION value.

    >>> ical_duration(datetime.timedelta(11))
    'P11D'
    >>> ical_duration(datetime.timedelta(-14))
    '-P14D'
    >>> ical_duration(datetime.timedelta(1, 7384))
    'P1DT2H3M4S'
    >>> ical_duration(datetime.timedelta(1, 7380))
    'P1DT2H3M'
    >>> ical_duration(datetime.timedelta(1, 7200))
    'P1DT2H'
    >>> ical_duration(datetime.timedelta(0, 7200))
    'PT2H'
    >>> ical_duration(datetime.timedelta(0, 7384))
    'PT2H3M4S'
    >>> ical_duration(datetime.timedelta(0, 184))
    'PT3M4S'
    >>> ical_duration(datetime.timedelta(0, 22))
    'PT22S'
    >>> ical_duration(datetime.timedelta(0, 3622))
    'PT1H0M22S'
    """
    sign = ""
    if value.days < 0:
        sign = "-"
    timepart = ""
    if value.seconds:
        timepart = "T"
        hours = value.seconds // 3600
        minutes = value.seconds % 3600 // 60
        seconds = value.seconds % 60
        if hours:
            timepart += "%dH" % hours
        if minutes or (hours and seconds):
            timepart += "%dM" % minutes
        if seconds:
            timepart += "%dS" % seconds
    if value.days == 0 and timepart:
        return "%sP%s" % (sign, timepart)
    else:
        return "%sP%dD%s" % (sign, abs(value.days), timepart)


def parse_date(value):
    """Parse iCalendar DATE value.  Returns a date instance.

    >>> parse_date('20030405')
    datetime.date(2003, 4, 5)
    >>> parse_date('20030405T060708')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar date: '20030405T060708'
    >>> parse_date('')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar date: ''
    >>> parse_date('yyyymmdd')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar date: 'yyyymmdd'
    """
    if len(value) != 8:
        raise ValueError('Invalid iCalendar date: %r' % value)
    try:
        y, m, d = int(value[0:4]), int(value[4:6]), int(value[6:8])
    except ValueError:
        raise ValueError('Invalid iCalendar date: %r' % value)
    else:
        return datetime.date(y, m, d)


def parse_date_time(value):
    """Parse iCalendar DATE-TIME value.  Returns a datetime instance.

    >>> parse_date_time('20030405T060708')
    datetime.datetime(2003, 4, 5, 6, 7, 8)
    >>> parse_date_time('20030405T060708Z')
    datetime.datetime(2003, 4, 5, 6, 7, 8)
    >>> parse_date_time('20030405T060708A')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar date-time: '20030405T060708A'
    >>> parse_date_time('')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar date-time: ''
    """
    datetime_rx = re.compile(r'(\d{4})(\d{2})(\d{2})'
                             r'T(\d{2})(\d{2})(\d{2})(Z?)$')
    match = datetime_rx.match(value)
    if match is None:
        raise ValueError('Invalid iCalendar date-time: %r' % value)
    y, m, d, hh, mm, ss, utc = match.groups()
    return datetime.datetime(int(y), int(m), int(d),
                             int(hh), int(mm), int(ss))


def parse_period(value):
    """Parse iCalendar PERIOD value.  Returns a Period instance.

    >>> p = parse_period('20030405T060708/20030405T060709')
    >>> print repr(p).replace('),', '),\\n      ')
    Period(datetime.datetime(2003, 4, 5, 6, 7, 8),
           datetime.datetime(2003, 4, 5, 6, 7, 9))
    >>> parse_period('20030405T060708/PT1H1M1S')
    Period(datetime.datetime(2003, 4, 5, 6, 7, 8), datetime.timedelta(0, 3661))
    >>> parse_period('xyzzy')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar period: 'xyzzy'
    >>> parse_period('foo/foe')
    Traceback (most recent call last):
      ...
    ValueError: Invalid iCalendar period: 'foo/foe'
    """
    parts = value.split('/')
    if len(parts) != 2:
        raise ValueError('Invalid iCalendar period: %r' % value)
    try:
        start = parse_date_time(parts[0])
        try:
            end_or_duration = parse_date_time(parts[1])
        except ValueError:
            end_or_duration = parse_duration(parts[1])
    except ValueError:
        raise ValueError('Invalid iCalendar period: %r' % value)
    else:
        return Period(start, end_or_duration)

def parse_attendee(value):
    """Parse list of attendees. Return list of person instances.

    >>> attendees = [('mailto:sidnei@awkly.org',
    ...              {'RSVP': 'FALSE',
    ...               'ROLE': 'REQ-PARTICIPANT',
    ...               'PARTSTAT': 'ACCEPTED',
    ...               'CN': 'Sidnei da Silva'}),
    ...             ('mailto:paul@zope-europe.org',
    ...              {'RSVP': 'FALSE',
    ...               'ROLE': 'REQ-PARTICIPANT',
    ...               'PARTSTAT': 'TENTATIVE',
    ...               'CN': 'Paul Everitt'})]
    >>> parse_attendee(attendees)
    [Person('sidnei@awkly.org', 'Sidnei da Silva'), Person('paul@zope-europe.org', 'Paul Everitt')]
    """
    res = []
    for k, v in value:
        res.append(Person(k, v.get('CN')))
    return res

def parse_categories(value):
    """Parse list of categories. Return list of category strings.

    >>> categories = [('testing',
    ...              {}),
    ...             ('another',
    ...              {})]
    >>> parse_categories(categories)
    ['testing', 'another']
    """
    return [k for k, v in value]

class Person:
    """ A person """

    def __init__(self, mail, name):
        mail = mail.lower()
        if not 'mailto:' in mail:
            mail = ''
        else:
            mail = filter(None, mail.split('mailto:'))
        self.mail = mail and mail[0] or ''
        self.name = name

    def __repr__(self):
        return "Person%r" % ((self.mail, self.name),)

    def __cmp__(self, other):
        if not isinstance(other, Person):
            raise NotImplementedError('Cannot compare Person with %r' % other)
        return cmp((self.mail, self.name),
                   (other.mail, other.name))

class Period:
    """A period of time"""

    def __init__(self, start, end_or_duration):
        self.start = start
        self.end_or_duration = end_or_duration
        if isinstance(end_or_duration, datetime.timedelta):
            self.duration = end_or_duration
            self.end = self.start + self.duration
        else:
            self.end = end_or_duration
            self.duration = self.end - self.start
        if self.start > self.end:
            raise ValueError("Start time is greater than end time")

    def __repr__(self):
        return "Period(%r, %r)" % (self.start, self.end_or_duration)

    def __cmp__(self, other):
        if not isinstance(other, Period):
            raise NotImplementedError('Cannot compare Period with %r' % other)
        return cmp((self.start, self.end), (other.start, other.end))

    def overlaps(self, other):
        if self.start > other.start:
            return other.overlaps(self)
        if self.start <= other.start < self.end:
            return True
        return False


class VEvent:
    """iCalendar event.

    Life cycle: when a VEvent is created, a number of properties should be
    added to it using the add method.  Then validate should be called.
    After that you can start using query methods (getOne, hasProp, iterDates).

    Events are classified into two kinds:
     - normal events
     - all-day events

    All-day events are identified by their DTSTART property having a DATE value
    instead of the default DATE-TIME.  All-day events should satisfy the
    following requirements (otherwise an exception will be raised):
     - DURATION property (if defined) should be an integral number of days
     - DTEND property (if defined) should have a DATE value
     - any RDATE and EXDATE properties should only contain DATE values

    The first two requirements are stated in RFC 2445; I'm not so sure about
    the third one.
    """

    default_type = {
        # Default value types for some properties
        'DTSTAMP': 'DATE-TIME',
        'DTSTART': 'DATE-TIME',
        'CREATED': 'DATE-TIME',
        'DTEND': 'DATE-TIME',
        'DURATION': 'DURATION',
        'LAST-MODIFIED': 'DATE-TIME',
        'PRIORITY': 'INTEGER',
        'RECURRENCE-ID': 'DATE-TIME',
        'SEQUENCE': 'INTEGER',
        'URL': 'URI',
        'ATTACH': 'URI',
        'EXDATE': 'DATE-TIME',
        'EXRULE': 'RECUR',
        'RDATE': 'DATE-TIME',
        'RRULE': 'RECUR',
        'LOCATION': 'TEXT',
        'UID': 'TEXT',
        'ATTENDEE':'ATTENDEE',
        }

    converters = {
        'INTEGER': int,
        'DATE': parse_date,
        'DATE-TIME': parse_date_time,
        'DURATION': parse_duration,
        'PERIOD': parse_period,
        'TEXT': parse_text,
        }

    complex_converters = {
        'ATTENDEE': parse_attendee,
        'CATEGORIES': parse_categories,
        }

    singleton_properties = Set([
        'DTSTAMP',
        'DTSTART',
        'UID',
        'CLASS',
        'CREATED',
        'DESCRIPTION',
        'DTEND',
        'DURATION',
        'GEO',
        'LAST-MODIFIED',
        'LOCATION',
        'ORGANIZER',
        'PRIORITY',
        'RECURRENCE-ID',
        'SEQUENCE',
        'STATUS',
        'SUMMARY',
        'TRANSP',
        'URL',
        ])

    rdate_types = Set(['DATE', 'DATE-TIME', 'PERIOD'])
    exdate_types = Set(['DATE', 'DATE-TIME'])

    def __init__(self):
        self._props = {}

    def add(self, property, value, params=None):
        """Add a property.

        Property name is case insensitive.  Params should be a dict from
        uppercased parameter names to their values.

        Multiple calls to add with the same property name override the
        value.  This is sufficient for now, but will have to be changed
        soon.
        """
        if params is None:
            params = {}
        key = property.upper()
        if key in self._props:
            if key in self.singleton_properties:
                raise ICalParseError("Property %s cannot occur more than once"
                                     % key)
            self._props[key].append((value, params))
        else:
            self._props[key] = [(value, params)]

    def validate(self):
        """Check that this VEvent has all the necessary properties.

        Also sets the following attributes:
          uid               The unique id of this event
          summary           Textual summary of this event
          all_day_event     True if this is an all-day event
          dtstart           start of the event (inclusive)
          dtend             end of the event (not inclusive)
          duration          length of the event
          location          location of the event
          rdates            a list of recurrence dates or periods
          exdates           a list of exception dates
        """
        if not self.hasProp('UID'):
            raise ICalParseError("VEVENT must have a UID property")
        if not self.hasProp('DTSTART'):
            raise ICalParseError("VEVENT must have a DTSTART property")
        if self._getType('DTSTART') not in ('DATE', 'DATE-TIME'):
            raise ICalParseError("DTSTART property should have a DATE or"
                                 " DATE-TIME value")
        if self.hasProp('DTEND'):
            if self._getType('DTEND') != self._getType('DTSTART'):
                raise ICalParseError("DTEND property should have the same type"
                                     " as DTSTART")
            if self.hasProp('DURATION'):
                raise ICalParseError("VEVENT cannot have both a DTEND"
                                     " and a DURATION property")
        if self.hasProp('DURATION'):
            if self._getType('DURATION') != 'DURATION':
                raise ICalParseError("DURATION property should have type"
                                     " DURATION")

        self.uid = self.getOne('UID')
        self.summary = self.getOne('SUMMARY')

        self.all_day_event = self._getType('DTSTART') == 'DATE'
        self.dtstart = self.getOne('DTSTART')
        if self.hasProp('DURATION'):
            self.duration = self.getOne('DURATION')
            self.dtend = self.dtstart + self.duration
        else:
            self.dtend = self.getOne('DTEND', None)
            if self.dtend is None:
                self.dtend = self.dtstart
                if self.all_day_event:
                    self.dtend += datetime.date.resolution
            self.duration = self.dtend - self.dtstart

        self.location = self.getOne('LOCATION', None)

        if self.dtstart > self.dtend:
            raise ICalParseError("Event start time should precede end time")
        if self.all_day_event and self.dtstart == self.dtend:
            raise ICalParseError("Event start time should precede end time")

        self.rdates = self._extractListOfDates('RDATE', self.rdate_types,
                                               self.all_day_event)
        self.exdates = self._extractListOfDates('EXDATE', self.exdate_types,
                                                self.all_day_event)

    def _extractListOfDates(self, key, accepted_types, all_day_event):
        """Parse a comma separated list of values.

        If all_day_event is True, only accepts DATE values.  Otherwise accepts
        all value types listed in 'accepted_types'.
        """
        dates = []
        default_type = self.default_type[key]
        for value, params in self._props.get(key, []):
            value_type = params.get('VALUE', default_type)
            if value_type not in accepted_types:
                raise ICalParseError('Invalid value type for %s: %s'
                                     % (key, value_type))
            if all_day_event and value_type != 'DATE':
                raise ICalParseError('I do not understand how to interpret '
                                     '%s values in %s for all-day events.'
                                     % (value_type, key))
            converter = self.converters.get(value_type)
            dates.extend(map(converter, value.split(',')))
        return dates

    def _getType(self, property):
        """Return the type of the property value."""
        key = property.upper()
        values = self._props.get(key)
        default_type = self.default_type.get(key, 'TEXT')
        if values is None:
            return default_type
        if key in self.singleton_properties:
            assert len(values) == 1, (key, values)
        value, params = values[0]
        return params.get('VALUE', default_type)

    def getOne(self, property, default=None):
        """Return the value of a property as an appropriate Python object.

        Only call getOne for properties that do not occur more than once.
        """
        try:
            values = self._props[property.upper()]
            assert len(values) == 1
            value, params = values[0]
        except KeyError:
            return default
        else:
            converter = self.converters.get(self._getType(property))
            if converter is None:
                return value
            else:
                return converter(value)

    def getProp(self, property, default=None):
        """Return the value of a property as an appropriate Python object.
        """
        key = property.upper()
        try:
            values = self._props[key]
            if key in self.singleton_properties:
                assert len(values) == 1, (key, values)
        except KeyError:
            return default
        else:
            res = []
            for val, params in values:
                c_params = {}
                v_conv = self.converters.get(self._getType(key))
                if v_conv is not None:
                    val = v_conv(val)
                res.append((val, c_params))
                for k, v in params.items():
                    p_key = k.upper()
                    p_conv = self.converters.get(self._getType(p_key))
                    if p_conv is not None:
                        v = p_conv(v)
                    c_params[k] = v

            if key not in self.singleton_properties:
                c_conv = self.complex_converters.get(key)
                if c_conv is not None:
                    return c_conv(res)
            else:
                return res[0][0]
            return res

    def hasProp(self, property):
        """Return True if this VEvent has a named property."""
        return property.upper() in self._props

    def iterDates(self):
        """Iterate over all dates within this event.

        This is only valid for all-day events at the moment.
        """
        if not self.all_day_event:
            raise ValueError('iterDates is only defined for all-day events')

        # Find out the set of start dates
        start_set = {self.dtstart: None}
        for rdate in self.rdates:
            start_set[rdate] = rdate
        for exdate in self.exdates:
            if exdate in start_set:
                del start_set[exdate]

        # Find out the set of all dates
        date_set = Set(start_set)
        duration = self.duration.days
        for d in start_set:
            for n in range(1, duration):
                d += datetime.date.resolution
                date_set.add(d)

        # Yield all dates in chronological order
        dates = list(date_set)
        dates.sort()
        for d in dates:
            yield d


class ICalReader:
    """An object which reads in an iCal.

    Short grammar of iCalendar files (RFC 2445 is the full spec):

      contentline        = name *(";" param ) ":" value CRLF
        ; content line first must be unfolded by replacing CRLF followed by a
        ; single WSP with an empty string
      name               = x-name / iana-token
      x-name             = "X-" [vendorid "-"] 1*(ALPHA / DIGIT / "-")
      iana-token         = 1*(ALPHA / DIGIT / "-")
      vendorid           = 3*(ALPHA / DIGIT)
      param              = param-name "=" param-value *("," param-value)
      param-name         = iana-token / x-token
      param-value        = paramtext / quoted-string
      paramtext          = *SAFE-CHAR
      value              = *VALUE-CHAR
      quoted-string      = DQUOTE *QSAFE-CHAR DQUOTE

      NON-US-ASCII       = %x80-F8
      QSAFE-CHAR         = WSP / %x21 / %x23-7E / NON-US-ASCII
                         ; Any character except CTLs and DQUOTE
      SAFE-CHAR          = WSP / %x21 / %x23-2B / %x2D-39 / %x3C-7E
                          / NON-US-ASCII
                         ; Any character except CTLs, DQUOTE, ";", ":", ","
      VALUE-CHAR         = WSP / %x21-7E / NON-US-ASCII  ; anything except CTLs
      CR                 = %x0D
      LF                 = %x0A
      CRLF               = CR LF
      CTL                = %x00-08 / %x0A-1F / %x7F
      ALPHA              = %x41-5A / %x61-7A             ; A-Z / a-z
      DIGIT              = %x30-39                       ; 0-9
      DQUOTE             = %x22                          ; Quotation Mark
      WSP                = SPACE / HTAB
      SPACE              = %x20
      HTAB               = %x09
    """

    def __init__(self, file):
        self.file = file

    def _parseRow(record_str):
        """Parse a single content line.

        A content line consists of a property name (optionally followed by a
        number of parameters) and a value, separated by a colon.  Parameters
        (if present) are separated from the property name and from each other
        with semicolons.  Parameters are of the form name=value; value
        can be double-quoted.

        Returns a tuple (name, value, param_dict).  Case-insensitive values
        (i.e. property names, parameter names, unquoted parameter values) are
        uppercased.

        Raises ICalParseError on syntax errors.

        >>> ICalReader._parseRow('foo:bar')
        ('FOO', 'bar', {})
        >>> ICalReader._parseRow('foo;value=bar:BAZFOO')
        ('FOO', 'BAZFOO', {'VALUE': 'BAR'})
        """

        it = iter(record_str)
        getChar = it.next

        def err(msg):
            raise ICalParseError("%s in line:\n%s" % (msg, record_str))

        try:
            c = getChar()
            # name
            key = ''
            while c.isalnum() or c == '-':
                key += c
                c = getChar()
            if not key:
                err("Missing property name")
            key = key.upper()
            # optional parameters
            params = {}
            while c == ';':
                c = getChar()
                # param name
                param = ''
                while c.isalnum() or c == '-':
                    param += c
                    c = getChar()
                if not param:
                    err("Missing parameter name")
                param = param.upper()
                # =
                if c != '=':
                    err("Expected '='")
                # value (or rather a list of values)
                pvalues = []
                while True:
                    c = getChar()
                    if c == '"':
                        c = getChar()
                        pvalue = ''
                        while c >= ' ' and c not in ('\177', '"'):
                            pvalue += c
                            c = getChar()
                        # value is case-sensitive in this case
                        if c != '"':
                            err("Expected '\"'")
                        c = getChar()
                    else:
                        # unquoted value
                        pvalue = ''
                        while c >= ' ' and c not in ('\177', '"', ';', ':',
                                                     ','):
                            pvalue += c
                            c = getChar()
                        pvalue = pvalue.upper()
                    pvalues.append(pvalue)
                    if c != ',':
                        break
                if len(pvalues) > 1:
                    params[param] = pvalues
                else:
                    params[param] = pvalues[0]
            # colon and value
            if c != ':':
                err("Expected ':'")
            value = ''.join(it)
        except StopIteration:
            err("Syntax error")
        else:
            return (key, value, params)

    _parseRow = staticmethod(_parseRow)

    def _iterRow(self):
        """A generator that returns one record at a time, as a tuple of
        (name, value, params).
        """
        record = []
        for line in self.file.readlines():
            if not line.strip():
                continue
            if line[0] in '\t ':
                line = line[1:]
            elif record:
                yield self._parseRow("".join(record))
                record = []
            if line.endswith('\r\n'):
                record.append(line[:-2])
            elif line.endswith('\n'):
                # strictly speaking this is a violation of RFC 2445
                record.append(line[:-1])
            else:
                # strictly speaking this is a violation of RFC 2445
                record.append(line)
        if record:
            yield self._parseRow("".join(record))

    def iterEvents(self):
        """Iterate over all VEVENT objects in an ICalendar file."""
        iterator = self._iterRow()

        # Check that the stream begins with BEGIN:VCALENDAR
        try:
            key, value, params = iterator.next()
            if (key, value, params) != ('BEGIN', 'VCALENDAR', {}):
                raise ICalParseError('This is not iCalendar')
        except StopIteration:
            # The file is empty.  Mozilla produces a 0-length file when
            # publishing an empty calendar.  Let's accept it as a valid
            # calendar that has no events.  XXX I'm not sure if a 0-length
            # file is a valid text/calendar object according to RFC 2445.
            raise
        component_stack = ['VCALENDAR']

        # Extract all VEVENT components
        obj = None
        for key, value, params in iterator:
            if key == "BEGIN":
                if obj is not None:
                    # Subcomponents terminate the processing of a VEVENT
                    # component.  We can get away with this now, because we're
                    # not interested in alarms and RFC 2445 specifies, that all
                    # properties inside a VEVENT component ought to precede any
                    # VALARM subcomponents.
                    obj.validate()
                    yield obj
                    obj = None
                if not component_stack and value != "VCALENDAR":
                    raise ICalParseError("Text outside VCALENDAR component")
                if value == "VEVENT":
                    obj = VEvent()
                component_stack.append(value)
            elif key == "END":
                if obj is not None and value == "VEVENT":
                    obj.validate()
                    yield obj
                    obj = None
                if not component_stack or component_stack[-1] != value:
                    raise ICalParseError("Mismatched BEGIN/END")
                component_stack.pop()
            elif obj is not None:
                obj.add(key, value, params)
            elif not component_stack:
                raise ICalParseError("Text outside VCALENDAR component")
        if component_stack:
            raise ICalParseError("Unterminated components")

class CalendarEvent(Persistent):

    def __init__(self, dt, duration, title, owner=None, context=None):

        self.dtstart = dt
        self.duration = duration
        self.title = title
        self.owner = owner
        self.context = context

    def __cmp__(self, other):
        if not isinstance(other, CalendarEvent):
            raise NotImplementedError('Cannot compare CalendarEvent with %r'
                                    % other)
        return cmp((self.dtstart, self.title, self.duration,
                    hash(self.owner), hash(self.context)),
                (other.dtstart, other.title, other.duration,
                    hash(other.owner), hash(other.context)))

    def __hash__(self):
        return hash((self.dtstart, self.title, self.duration,
                    self.owner, self.context))

    def __repr__(self):
        return ("CalendarEvent%r"
                % ((self.dtstart, self.duration, self.title,
                    self.owner, self.context), ))

class Calendar(Persistent):

    def __init__(self):
        self._events = Set()
        self.__name__ = None
        self.__parent__ = None

    def __iter__(self):
        return iter(self._events)


    def __getitem__(self, idx):
        return list(self._events)[idx]

    def byDate(self, date):
        cal = Calendar()
        for event in self:
            event_start = event.dtstart.date()
            event_end = (event.dtstart + event.duration).date()
            if event_start <= date and event_end >= date:
                cal.addEvent(event)
        return cal

    def addEvent(self, event):
        self._events.add(event)
        self._events = self._events  # make persistence work

    def removeEvent(self, event):
        self._events.remove(event)
        self._events = self._events  # make persistence work

    def update(self, calendar):
        for event in calendar:
            self._events.add(event)
        self._events = self._events  # make persistence work

    def clear(self):
        self._events.clear()
        self._events = self._events  # make persistence work



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Some portions of this module are Copyright Shuttleworth Foundation.
# The original copyright statement is reproduced below.
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2003 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Common utilities (stubs, mixins) for schooltool unit tests.

$Id: utils.py,v 1.2 2004/09/17 13:58:26 dreamcatcher Exp $
"""

import sys
import unittest
from pprint import pformat
from Products.Calendaring.tests.helpers import normalize_xml, diff

__metaclass__ = type


class EqualsSortedMixin:
    """Mixin that adds a helper method for comparing lists ignoring order."""

    def assertEqualsSorted(self, a, b):
        x = list(a)
        y = list(b)
        x.sort()
        y.sort()
        self.assertEquals(x, y)

    assertEqualSorted = assertEqualsSorted


class NiceDiffsMixin:
    """Mixin that changes assertEquals to show a unified diff of pretty-printed
    values.
    """

    def assertEquals(self, results, expected, msg=None):
        if msg is None:
            msg = "\n" + diff(pformat(expected), pformat(results))
        unittest.TestCase.assertEquals(self, results, expected, msg)

    assertEqual = assertEquals


class XMLCompareMixin:

    def assertEqualsXML(self, result, expected, recursively_sort=()):
        """Assert that two XML documents are equivalent.

        If recursively_sort is given, it is a sequence of tags that
        will have test:sort="recursively" appended to their attribute lists
        in 'result' text.  See the docstring for normalize_xml for more
        information about this attribute.
        """
        result = normalize_xml(result, recursively_sort=recursively_sort)
        expected = normalize_xml(expected, recursively_sort=recursively_sort)
        self.assertEquals(result, expected, "\n" + diff(expected, result))

    assertEqualXML = assertEqualsXML


class QuietLibxml2Mixin:
    """Text mixin that disables libxml2 error reporting.

    Sadly the API of libxml2 does not allow us to restore the error reporting
    function in tearDown.  <Insert derogatory comments here>
    """

    def setUpLibxml2(self):
        import libxml2
        libxml2.registerErrorHandler(lambda ctx, error: None, None)

    def tearDownLibxml2(self):
        import libxml2
        # It's not possible to restore the error handler that was installed
        # before (libxml2 API limitation), so we set up a generic one that
        # prints everything to stdout.
        def on_error_callback(ctx, msg):
            sys.stderr.write(msg)
        libxml2.registerErrorHandler(on_error_callback, None)


from Products.CMFCore.utils import getToolByName
from Products.Archetypes.ArchetypeTool import FactoryTypeInformation
from Products.Archetypes.ArchetypeTool import fixActionsForType, getType
# Mostly copied from ArchetypeTool.manage_installType
def installType(self, typeName, package, portal_type=None, ignore_actions=False):
    typesTool = getToolByName(self, 'portal_types')

    typeinfo_name="%s: %s" % (package, typeName)
    typeDesc = getType(typeName, package)
    if not portal_type:
        portal_type = typeDesc['portal_type']

    try:
        typesTool._delObject(portal_type)
    except:
        pass

    typesTool.manage_addTypeInformation(
        FactoryTypeInformation.meta_type,
        id=portal_type,
        typeinfo_name=typeinfo_name)
    t = getattr(typesTool, typeName, None)
    if t:
        t.title = getattr(typeDesc['klass'], 'archetype_name',
                          portal_type)

    if not ignore_actions:
        # and update the actions as needed
        fixActionsForType(typeDesc['klass'], typesTool)



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: day.py,v 1.2 2004/09/16 13:46:09 dreamcatcher Exp $
"""

from dateutil.relativedelta import relativedelta
from Products.Calendaring.model import ModelView, registerView

class OneDayView(ModelView):
    """ A view for a single day """

    unit = 'day'
    _sub = 'hour_view'
    name = 'day_view'
    _args = 3
    _leap = relativedelta(0)

registerView(OneDayView)



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: test_model.py,v 1.3 2004/09/16 13:46:09 dreamcatcher Exp $
"""
import os, sys

if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

import unittest
from doctest import DocTestSuite
from Testing import ZopeTestCase

# Install our product
ZopeTestCase.installProduct('MimetypesRegistry')
ZopeTestCase.installProduct('PortalTransforms')
ZopeTestCase.installProduct('Archetypes')
ZopeTestCase.installProduct('Calendaring')

import datetime
from dateutil.relativedelta import relativedelta
import urllib, urlparse, cgi
from Products.Calendaring.model import SECOND, DAY
from Products.Calendaring.model.year import OneYearView
from Products.Calendaring.model.month import OneMonthView
from Products.Calendaring.model.month import ThreeMonthsView
from Products.Calendaring.model.week import OneWeekView
from Products.Calendaring.model.week import ThreeWeeksView
from Products.Calendaring.model.week import WorkdaysView
from Products.Calendaring.model.day import OneDayView
from Products.Calendaring.model.hour import OneHourView
from Products.Calendaring.model import makeRange
from OFS.SimpleItem import SimpleItem

START_DATE = datetime.datetime(2004, 4, 27, 9, 5, 40)

SUBS = {OneYearView:(12, OneMonthView),
        OneMonthView:(5, OneWeekView),
        ThreeMonthsView:(3, OneMonthView),
        OneWeekView:(7, OneDayView),
        ThreeWeeksView:(3, OneWeekView),
        WorkdaysView:(5, OneDayView),
        OneDayView:(24, OneHourView),
        OneHourView:(1, OneHourView)}

def compareURL(url, other):
    parsed_url = list(urlparse.urlparse(url))
    # get the existing query string
    qs = parsed_url[4]
    # parse the query into a dict
    first = cgi.parse_qs(qs, 1)
    # now, the same with the other
    parsed_url = list(urlparse.urlparse(other))
    # get the existing query string
    qs = parsed_url[4]
    # parse the query into a dict
    second = cgi.parse_qs(qs, 1)
    for k, v in first.items():
        if second.get(k) != v:
            return False
    return True

def getSub(instance):
    return SUBS[instance.__class__][1]

def getSubCount(instance):
    return SUBS[instance.__class__][0]

class FakeRequest:

    def __init__(self, env={}):
        self.other = {}
        self.form = {}
        self.__dict__.update(env)

class TestMakeRange(unittest.TestCase):

    def _range(self, date, unit):
        return makeRange(date, unit)

    def test_year(self):
        r = self._range
        self.assertEquals(r(START_DATE, 'year'),
                          (datetime.datetime(2004, 1, 1, 0, 0, 0),
                           datetime.datetime(2005, 1, 1, 0, 0, 0),
                           relativedelta(years=1)))
        self.assertEquals(r(datetime.datetime(2005, 1, 1, 0, 0, 0), 'year'),
                          (datetime.datetime(2005, 1, 1, 0, 0, 0),
                           datetime.datetime(2006, 1, 1, 0, 0, 0),
                           relativedelta(years=1)))

    def test_month(self):
        r = self._range
        self.assertEquals(r(START_DATE, 'month'),
                          (datetime.datetime(2004, 4, 1, 0, 0, 0),
                           datetime.datetime(2004, 5, 1, 0, 0, 0),
                           relativedelta(months=1)))
        self.assertEquals(r(datetime.datetime(2004, 5, 1, 0, 0, 0), 'month'),
                          (datetime.datetime(2004, 5, 1, 0, 0, 0),
                           datetime.datetime(2004, 6, 1, 0, 0, 0),
                           relativedelta(months=1)))

    def test_3month(self):
        r = self._range
        self.assertEquals(r(START_DATE, '3month'),
                          (datetime.datetime(2004, 3, 1, 0, 0, 0),
                           datetime.datetime(2004, 6, 1, 0, 0, 0),
                           relativedelta(months=3)))
        self.assertEquals(r(datetime.datetime(2004, 6, 1, 0, 0, 0), '3month'),
                          (datetime.datetime(2004, 5, 1, 0, 0, 0),
                           datetime.datetime(2004, 8, 1, 0, 0, 0),
                           relativedelta(months=3)))

    def test_week(self):
        r = self._range
        self.assertEquals(r(START_DATE, 'week'),
                          (datetime.datetime(2004, 4, 26, 0, 0, 0),
                           datetime.datetime(2004, 5, 3, 0, 0, 0),
                           relativedelta(days=7)))
        self.assertEquals(r(datetime.datetime(2004, 5, 2, 0, 0, 0), 'week'),
                          (datetime.datetime(2004, 4, 26, 0, 0, 0),
                           datetime.datetime(2004, 5, 3, 0, 0, 0),
                           relativedelta(days=7)))
        self.assertEquals(r(datetime.datetime(2004, 5, 3, 0, 0, 0), 'week'),
                          (datetime.datetime(2004, 5, 3, 0, 0, 0),
                           datetime.datetime(2004, 5, 10, 0, 0, 0),
                           relativedelta(days=7)))

    def test_3week(self):
        r = self._range
        self.assertEquals(r(START_DATE, '3week'),
                          (datetime.datetime(2004, 4, 19, 0, 0, 0),
                           datetime.datetime(2004, 5, 10, 0, 0, 0),
                           relativedelta(days=21)))
        self.assertEquals(r(datetime.datetime(2004, 5, 2, 0, 0, 0), '3week'),
                          (datetime.datetime(2004, 4, 19, 0, 0, 0),
                           datetime.datetime(2004, 5, 10, 0, 0, 0),
                           relativedelta(days=21)))
        self.assertEquals(r(datetime.datetime(2004, 5, 10, 0, 0, 0), '3week'),
                          (datetime.datetime(2004, 5, 3, 0, 0, 0),
                           datetime.datetime(2004, 5, 24, 0, 0, 0),
                           relativedelta(days=21)))

class TestModel(unittest.TestCase):

    factory = None

    def setUp(self):
        req = FakeRequest(env={'URL1':''})
        self.root = SimpleItem()
        self.root.REQUEST = req
        self.factory._factory_hook = getSub
        self.view = self.factory(START_DATE, START_DATE)
        self.view = self.view.__of__(self.root)

    def test_next_previous(self):
        view = self.view
        self.failUnless(isinstance(view.next(), self.factory))
        self.failUnless(isinstance(view.previous(), self.factory))
        n = view.next()
        p = view.previous()
        self.assertEquals(n.start, view.end, (view, n))
        self.assertEquals(n.end, n.start + n.resolution, n)
        self.assertEquals(p.end, view.start, (view, p))
        self.assertEquals(p.start, p.end - p.resolution, p)

    def test_sub(self):
        view = self.view
        self.assertEquals(view.sub_factory, getSub(view))
        subs = list(view.parts)
        self.assertEquals(len(subs), getSubCount(view))
        self.failUnless(isinstance(subs[0], getSub(view)))

    def test_current(self):
        view = self.view
        next = view.next()
        previous = view.previous()
        self.assertEquals(view.today, START_DATE)
        self.failUnless(view.current, view)
        self.assertEquals(next.today, START_DATE)
        self.failIf(next.current, next)
        self.assertEquals(previous.today, START_DATE)
        self.failIf(previous.current, previous)

class TestOneYearModel(TestModel):

    factory = OneYearView

    def test_next_previous_url(self):
        view = self.view
        next = view.next()
        previous = view.previous()
        got = view.next_url()
        expected = '%s?cyear=%s' % (view.name, next.start.year)
        self.failUnless(
            compareURL(got, expected), (got, expected))
        got = view.previous_url()
        expected = '%s?cyear=%s' % (view.name, previous.start.year)
        self.failUnless(
            compareURL(got, expected), (got, expected))

class TestMonthsModel(TestModel):

    factory = OneMonthView

    def test_next_previous_url(self):
        view = self.view
        next = view.next()
        previous = view.previous()
        got = view.next_url()
        expected = '%s?cyear=%s&cmonth=%s' % (view.name, next.start.year,
                                              next.start.month)
        self.failUnless(
            compareURL(got, expected), (got, expected))
        got = view.previous_url()
        expected = '%s?cyear=%s&cmonth=%s' % (view.name, previous.start.year,
                                              previous.start.month)
        self.failUnless(
            compareURL(got, expected), (got, expected))

class TestThreeMonthsModel(TestMonthsModel):

    factory = ThreeMonthsView

class TestDaysModel(TestModel):

    factory = OneDayView

    def test_next_previous_url(self):
        view = self.view
        next = view.next()
        previous = view.previous()
        got = view.next_url()
        expected = '%s?cyear=%s&cmonth=%s&cday=%s' % (view.name,
                                                   next.start.year,
                                                   next.start.month,
                                                   next.start.day)
        self.failUnless(
            compareURL(got, expected), (got, expected))
        got = view.previous_url()
        expected = '%s?cyear=%s&cmonth=%s&cday=%s' % (view.name,
                                                   previous.start.year,
                                                   previous.start.month,
                                                   previous.start.day)
        self.failUnless(
            compareURL(got, expected), (got, expected))

class TestWorkdaysModel(TestDaysModel):

    factory = WorkdaysView

    def test_next_previous(self):
        view = self.view
        self.failUnless(isinstance(view.next(), self.factory))
        self.failUnless(isinstance(view.previous(), self.factory))
        n = view.next()
        p = view.previous()
        self.assertEquals(n.start, view.end + datetime.timedelta(2), (view, n))
        self.assertEquals(n.end, n.start + n.resolution, n)
        self.assertEquals(p.end + datetime.timedelta(2), view.start, (view, p))
        self.assertEquals(p.start, p.end - p.resolution, p)

class TestOneWeekModel(TestDaysModel):

    factory = OneWeekView

class TestThreeWeeksModel(TestDaysModel):

    factory = ThreeWeeksView


class TestHoursModel(TestModel):

    factory = OneHourView

    def test_next_previous_url(self):
        view = self.view
        next = view.next()
        previous = view.previous()
        got = view.next_url()
        expected = '%s?cyear=%s&cmonth=%s&cday=%s&hour=%s' % (view.name,
                                                           next.start.year,
                                                           next.start.month,
                                                           next.start.day,
                                                           next.start.hour)
        self.failUnless(
            compareURL(got, expected), (got, expected))
        got = view.previous_url()
        expected = '%s?cyear=%s&cmonth=%s&cday=%s&hour=%s' % (view.name,
                                                           previous.start.year,
                                                           previous.start.month,
                                                           previous.start.day,
                                                           previous.start.hour)
        self.failUnless(
            compareURL(got, expected), (got, expected))


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestMakeRange))
    suite.addTest(unittest.makeSuite(TestOneYearModel))
    suite.addTest(unittest.makeSuite(TestMonthsModel))
    suite.addTest(unittest.makeSuite(TestThreeMonthsModel))
    suite.addTest(unittest.makeSuite(TestDaysModel))
    suite.addTest(unittest.makeSuite(TestWorkdaysModel))
    suite.addTest(unittest.makeSuite(TestOneWeekModel))
    suite.addTest(unittest.makeSuite(TestThreeWeeksModel))
    suite.addTest(unittest.makeSuite(TestHoursModel))
    return suite

if __name__ == '__main__':
    framework(descriptions=1, verbosity=1)



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: test_traversal.py,v 1.6 2005/01/25 01:03:48 dreamcatcher Exp $
"""
import os, sys
from os import curdir
from os.path import join, abspath, dirname, split
from StringIO import StringIO

try:
    __file__
except NameError:
    # Test was called directly, so no __file__ global exists.
    _prefix = abspath(curdir)
else:
    # Test was called by another test.
    _prefix = abspath(dirname(__file__))

if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

# Load fixture
import unittest
from Testing import ZopeTestCase

# Install our product
ZopeTestCase.installProduct('MimetypesRegistry')
ZopeTestCase.installProduct('PortalTransforms')
ZopeTestCase.installProduct('Archetypes')
ZopeTestCase.installProduct('Calendaring')
ZopeTestCase.installProduct('ATContentTypes')

import datetime
from DateTime import DateTime
from Products.CMFPlone.tests import PloneTestCase
from Products.Calendaring.tests.utils import installType

class TestViewTraversal(PloneTestCase.PloneTestCase):

    def _setupCalendar(self):
        self.qi = self.portal.portal_quickinstaller
        self.qi.installProduct('PortalTransforms')
        self.qi.installProduct('Archetypes')
        self.qi.installProduct('Calendaring')
        self.ct = self.portal.portal_calendar
        self.wf = self.portal.portal_workflow

        # Install ATEvent as Event
        installType(self.portal,
                    'ATEvent',
                    'ATContentTypes',
                    'Event',
                    ignore_actions=True)

        self.ct.setMapping('atct')
        self.portal._v_skindata = None
        self.portal.setupCurrentSkin()

    def afterSetUp(self):
        self.loginPortalOwner()
        self._setupCalendar()
        self.cat = self.portal.portal_catalog
        self.mt = self.portal.portal_membership
        self.wt = self.portal.portal_workflow
        self.portal.invokeFactory('Calendar', id='test_calendar')
        self.cal = self.portal.test_calendar
        self.file = open(join(_prefix, 'data', 'tasks.ics'), 'r')
        self.items = self.cal.fromFile(self.file)
        self.file.close()

    def _check_day_view(self, dt, view, evcount=(1, 1)):
        self.assertEquals(view.start, dt)
        self.assertEquals(view.end, dt + view.resolution)
        gef = self.ct.getEventsForThisDay
        self.assertEquals(len(gef(DateTime(2004, 4, 26))), 1)
        parts = list(view.parts)
        self.assertEquals(len(parts), 24)
        hour = parts[9]
        events = list(hour.events)
        self.assertEquals(len(events), evcount[0])
        if evcount[0]:
            ev0 = events[0].getObject()
            review_state = self.wf.getInfoFor(ev0, 'review_state')
            self.assertEquals(review_state, 'private')
        hour = parts[15]
        events = list(hour.events)
        self.assertEquals(len(events), evcount[1])
        if evcount[1]:
            ev0 = events[0].getObject()
            review_state = self.wf.getInfoFor(ev0, 'review_state')
            self.assertEquals(review_state, 'published')

    def test_traversal_day_view(self):
        cal = self.cal
        ct = self.ct
        uid = cal.UID()
        request = cal.REQUEST
        dt = datetime.datetime(2004, 4, 26)
        request.set('cyear', 2004)
        request.set('cmonth', 4)
        request.set('cday', 26)
        path = '%s/day_view' % uid
        view = ct.restrictedTraverse(path)
        self._check_day_view(dt, view)

    def test_anon_traversal_day_view(self):
        self.logout()
        cal = self.cal
        ct = self.ct
        uid = cal.UID()
        request = cal.REQUEST
        dt = datetime.datetime(2004, 4, 26)
        request.set('cyear', 2004)
        request.set('cmonth', 4)
        request.set('cday', 26)
        path = '%s/day_view' % uid
        view = ct.restrictedTraverse(path)
        self._check_day_view(dt, view, evcount=(0, 1))

    def test_traversal_calendar_day_view(self):
        cal = self.cal
        request = cal.REQUEST
        dt = datetime.datetime(2004, 4, 26)
        request.set('cyear', 2004)
        request.set('cmonth', 4)
        request.set('cday', 26)
        path = 'day_view'
        view = cal.restrictedTraverse(path)
        self._check_day_view(dt, view)

    def test_anon_traversal_calendar_day_view(self):
        self.logout()
        cal = self.cal
        request = cal.REQUEST
        dt = datetime.datetime(2004, 4, 26)
        request.set('cyear', 2004)
        request.set('cmonth', 4)
        request.set('cday', 26)
        path = 'day_view'
        view = cal.restrictedTraverse(path)
        self._check_day_view(dt, view, evcount=(0,1))

    def _check_month_view(self, dt, view):
        self.assertEquals(view.start, dt)
        self.assertEquals(view.end, dt + view.resolution)
        gef = self.ct.getEventsForThisDay
        self.assertEquals(len(gef(DateTime(2004, 4, 26))), 1)
        parts = list(view.parts)
        self.assertEquals(len(parts), 5)

    def test_traversal_month_view(self):
        cal = self.cal
        ct = self.ct
        uid = cal.UID()
        request = cal.REQUEST
        dt = datetime.datetime(2004, 4, 1)
        request.set('cyear', 2004)
        request.set('cmonth', 4)
        request.set('cday', 26)
        path = '%s/month_view' % uid
        view = ct.restrictedTraverse(path)
        self._check_month_view(dt, view)

    def test_traversal_calendar_month_view(self):
        cal = self.cal
        request = cal.REQUEST
        dt = datetime.datetime(2004, 4, 1)
        request.set('cyear', 2004)
        request.set('cmonth', 4)
        request.set('cday', 26)
        path = 'month_view'
        view = cal.restrictedTraverse(path)
        self._check_month_view(dt, view)

    def _check_week_view(self, dt, view):
        self.assertEquals(view.start, dt)
        self.assertEquals(view.end, dt + view.resolution)
        gef = self.ct.getEventsForThisDay
        self.assertEquals(len(gef(DateTime(2004, 4, 26))), 1)
        parts = list(view.parts)
        self.assertEquals(len(parts), 7)

    def test_traversal_week_view(self):
        cal = self.cal
        ct = self.ct
        uid = cal.UID()
        request = cal.REQUEST
        dt = datetime.datetime(2004, 4, 26)
        request.set('cyear', 2004)
        request.set('cmonth', 4)
        request.set('cday', 26)
        path = '%s/week_view' % uid
        view = ct.restrictedTraverse(path)
        self._check_week_view(dt, view)

    def test_traversal_calendar_week_view(self):
        cal = self.cal
        request = cal.REQUEST
        dt = datetime.datetime(2004, 4, 26)
        request.set('cyear', 2004)
        request.set('cmonth', 4)
        request.set('cday', 26)
        path = 'week_view'
        view = cal.restrictedTraverse(path)
        self._check_week_view(dt, view)

    def test_HEAD(self):
        cal = self.cal
        request = cal.REQUEST
        response = request.RESPONSE
        self.assertEquals(cal.HEAD(request, response).status, 200)

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestViewTraversal))
    return suite

if __name__ == '__main__':
    framework(descriptions=1, verbosity=1)



#     Calendaring is a simple CMF/Plone calendaring implementation.
#     Copyright (C) 2004 Enfold Systems
#
#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program; if not, write to the Free Software
#     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from Products.Archetypes.public import listTypes
from Products.Archetypes.Extensions.utils import installTypes, install_subskin
from Products.CMFCore.utils import getToolByName
from Products.Calendaring.config import *
from StringIO import StringIO

def install_ctr(self, out):
    ctr = getToolByName(self, 'content_type_registry')
    if 'calendar_mime' not in ctr.predicate_ids:
        ctr.addPredicate('calendar_mime', 'major_minor' )
        ctr.getPredicate('calendar_mime').edit('text', 'calendar')
        ctr.assignTypeName('calendar_mime', 'Calendar')
        ctr.reorderPredicate('calendar_mime', 0)
        out.write('Installed CTR Predicate for text/calendar -> Calendar.\n')
    else:
        out.write('Predicate for text/calendar -> Calendar was already installed.\n')

    if 'calendar_ext' not in ctr.predicate_ids:
        ctr.addPredicate('calendar_ext', 'extension' )
        ctr.getPredicate('calendar_ext').edit('ics')
        ctr.assignTypeName('calendar_ext', 'Calendar')
        ctr.reorderPredicate('calendar_ext', 1)
        out.write('Installed CTR Predicate for .ics -> Calendar.\n')
    else:
        out.write('Predicate for .ics -> Calendar was already installed.\n')

def install_actions(self, out):
    ai = getToolByName(self, 'portal_actionicons')
    for category, config in ACTION_ICONS.items():
        for icon_id, info in config.items():
            if ai.queryActionIcon(category, icon_id, None) is None:
                ai.addActionIcon(category, icon_id,
                                 info[0], info[1])
                out.write("Installed action icon "
                          "for %s.\n" % info[1])
            else:
                out.write(("Action Icon for %s "
                           "was already Installed.\n"
                           % info[1]))

def install_tools(self, product, tools, out):
    addTool = self.manage_addProduct[product].manage_addTool
    for id in TOOLIDS:
        if hasattr(self, id):
            self.manage_delObjects(ids=[id])
    for tool in tools:
        addTool(tool)
        out.write("Created tool: %s.\n" % tool)

def install_types(self, out, types, project):
    installTypes(self, out, types, project)

def install(self):
    from Products.Calendaring.tools.calendar import CalendarTool
    out = StringIO()

    install_types(self, out, listTypes(PROJECTNAME), PROJECTNAME)
    install_subskin(self, out, GLOBALS)
    install_tools(self, PROJECTNAME, [CalendarTool.meta_type], out)
    install_actions(self, out)
    install_ctr(self, out)

    out.write("Successfully installed %s." % PROJECTNAME)
    return out.getvalue()



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: _cal.py,v 1.2 2004/09/16 13:46:09 dreamcatcher Exp $
"""
import calendar
from dateutil.relativedelta import relativedelta
from datetime import timedelta, datetime, date

WEEK = relativedelta(days=7)
WEEKDAY = relativedelta(days=5)
DAY = relativedelta(days=1)
HOUR = relativedelta(hours=1)
MONTH = relativedelta(months=1)
YEAR = relativedelta(years=1)
SECOND = relativedelta(seconds=1)

def yearRange(dt):
    start = dt.replace(month=1, day=1, hour=0, minute=0, second=0)
    end = start + YEAR
    delta = YEAR
    return start, end, delta

def monthRange(dt):
    start = dt.replace(day=1, hour=0, minute=0, second=0)
    delta = MONTH
    end = start + delta
    return start, end, delta

def threeMonthRange(dt):
    previous = dt - MONTH
    start = previous.replace(day=1, hour=0, minute=0, second=0)
    next = dt + MONTH
    next = next.replace(day=1, hour=0, minute=0, second=0)
    end = next + timedelta(calendar.monthrange(next.year, next.month)[1])
    delta = relativedelta(end, start)
    return start, end, delta

def weekRange(dt):
    start = dt - timedelta(dt.weekday())
    end = dt + timedelta(7 - dt.weekday())
    start = start.replace(hour=0, minute=0, second=0)
    end = end.replace(hour=0, minute=0, second=0)
    delta = relativedelta(end, start)
    return start, end, delta

def workdaysRange(dt):
    start = dt - timedelta(dt.weekday())
    end = dt + timedelta(5 - dt.weekday())
    start = start.replace(hour=0, minute=0, second=0)
    end = end.replace(hour=0, minute=0, second=0)
    delta = relativedelta(end, start)
    return start, end, delta

def threeWeekRange(dt):
    start = (dt - timedelta(dt.weekday())) - WEEK
    end = (dt + timedelta(7 - dt.weekday())) + WEEK
    start = start.replace(hour=0, minute=0, second=0)
    end = end.replace(hour=0, minute=0, second=0)
    delta = relativedelta(end, start)
    return start, end, delta

def dayRange(dt):
    start = dt
    end = dt + DAY
    start = start.replace(hour=0, minute=0, second=0)
    end = end.replace(hour=0, minute=0, second=0)
    return start, end, DAY

def hourRange(dt):
    start = dt.replace(minute=0, second=0)
    end = dt.replace(minute=0, second=0) + HOUR
    return start, end, HOUR

RANGES = {'day':dayRange,
          'week':weekRange,
          '3week':threeWeekRange,
          'month':monthRange,
          '3month':threeMonthRange,
          'year':yearRange,
          'workdays':workdaysRange,
          'hour':hourRange}

def makeRange(dt, res):
    func = RANGES[res]
    return func(dt)



## Script (Python) "portlet_calendaring_search_helper"
##bind container=container
##bind context=context
##bind namespace=
##bind script=script
##bind subpath=traverse_subpath
##parameters=states
##title=
##
#in order to have plones queryCatalog using
#a list of states we must proviede a rather clumsy parameter list
if not states:
    states = ['published']
if same_type(states, ''):
    states = [states]
result = ""
for state in states:
    if result:
        result += '&'
    result += 'review_state:list=%s' % state

return result



## Controller Python Script "calendar_export"
##bind container=container
##bind context=context
##bind namespace=
##bind script=script
##bind subpath=traverse_subpath
##parameters=
##title=Export Calendar
##
# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: calendar_export.py,v 1.1.1.1 2004/06/19 14:53:36 dreamcatcher Exp $
"""

REQUEST = context.REQUEST
RESPONSE = REQUEST.RESPONSE
name = context.getId()
if not name.endswith('.ics'):
    name = '%s.ics' % name
RESPONSE.setHeader('content-disposition',
                   'attachment; filename=%s' % name)

return context.asCalendar(REQUEST=context.REQUEST)



# Calendaring is a simple CMF/Plone calendaring implementation.
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Some portions of this module are Copyright Shuttleworth Foundation.
# The original copyright statement is reproduced below.
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2003 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id: calendar.py,v 1.4 2005/01/25 01:03:49 dreamcatcher Exp $
"""
from Products.CMFCalendar.CalendarTool import calendar
from OFS.PropertyManager import PropertyManager

from datetime import datetime, date
from types import ListType
from rfc822 import parseaddr, dump_address_pair
from BTrees.OOBTree import OOSet

from Globals import InitializeClass
from Acquisition import aq_base
from AccessControl import ClassSecurityInfo, Unauthorized

from Products.CMFCore.utils import getToolByName
from Products.CMFCore.WorkflowCore import WorkflowException
from Products.CMFPlone.CalendarTool import CalendarTool as BaseTool
from Products.Calendaring.parser import ICalReader, \
     ICalParseError, ical_text, ical_duration, Person
from Products.Calendaring.common import dt2DT
from Products.Calendaring.model import getView
from Products.Calendaring.model.traversal import ModelViewMixin

UID_ATTR = '_event_uid'

def generateUID(event):
    uid = getattr(aq_base(event), UID_ATTR, None)
    if uid is not None:
        return uid
    id = event.getId()
    if id.endswith('RID'):
        return id
    h_event = abs(hash(repr(event)))
    h_id = abs(hash(id))
    return "%s-%s-RID" % (h_event, h_id)

def generateId(event):
    uid = event.getProp('uid', None)
    if uid:
        return uid
    title = event.summary.lower().replace(' ', '_')
    time = ''.join(tuple(map(str, event.dtstart.timetuple()[:5])))
    return '_'.join((title[:15], time))

def person2text(value):
    text = value.name
    if value.mail:
        text = dump_address_pair((text, value.mail))
    return text

def convert(value):
    if type(value) is not ListType:
        return _convert(value)
    return map(_convert, value)

def _convert(value):
    for t, c in converters.items():
        if isinstance(value, t):
            return c(value)
    return value


status_mapping = {'TENTATIVE':('submit', 'reject', 'retract', 'show'),
                  'CONFIRMED':('publish',),
                  'CANCELLED':('hide',),
                  }
re_status_mapping = {'published':'CONFIRMED',
                     'pending':'TENTATIVE',
                     'visible':'TENTATIVE',
                     'private':'CANCELLED'}

converters = {datetime : dt2DT,
              date : dt2DT,
              Person : person2text}


kw_mappings = {'cmf':{'dtstart':'start_date',
                      'dtend':'end_date',
                      'summary':'title',
                      'description':'description'},
               'atct':{'dtstart':'startDate',
                       'dtend':'endDate',
                       'summary':'title',
                       'description':'description',
                       'location':'location',
                       'attendee':'attendees',
                       'categories':'subject'}
               }

class CalendarTool(ModelViewMixin, BaseTool, PropertyManager):
    """ A Calendar Tool implementation
    for import/export of events.
    """
    security = ClassSecurityInfo()

    _kw_map = kw_mappings['atct']
    event_type = 'Event'
    _properties = getattr(BaseTool, '_properties', ()) + (
        {'id':'event_type', 'type': 'string', 'mode':'w'},
        {'id':'published_states', 'type': 'lines', 'mode':'w'},)

    manage_options = BaseTool.manage_options + PropertyManager.manage_options

    def __init__(self):
        if hasattr(BaseTool, '__init__'):
            # Plone 2.1 doesn't seem to have a __init__ anymore.
            BaseTool.__init__(self)
        self._registry = OOSet()

    def manage_afterAdd(self, item, container):
        item.manage_changeProperties({'published_states' : ['published']})

    security.declarePrivate('setMapping')
    def setMapping(self, name):
        self._kw_map = kw_mappings[name]

    security.declarePublic('getDestination')
    def getDestination(self):
        mt = getToolByName(self, 'portal_membership')
        return mt.getHomeFolder()

    security.declarePublic('importCalendar')
    def importCalendar(self, stream, dest=None, kw_map=None, do_action=False):
        items = []
        mt = getToolByName(self, 'portal_membership')
        if mt.isAnonymousUser():
            raise Unauthorized
        wt = getToolByName(self, 'portal_workflow')
        if dest is None:
            dest = self.getDestination()
        reader = ICalReader(stream)
        if kw_map is None:
            kw_map = self._kw_map
        for ev in reader.iterEvents():
            id = generateId(ev)
            info = {'id': id}
            for k, v in kw_map.items():
                value = getattr(ev, k, ev.getProp(k, default=None))
                if value is None:
                    continue
                value = convert(value)
                info[v] = value

            if id in dest.objectIds():
                obj = dest._getOb(id)
                obj.update(**info)
            else:
                dest.invokeFactory(self.event_type, **info)
                obj = dest._getOb(id)
            uid = getattr(ev, 'uid', None)
            if uid is None:
                uid = generateUID(obj)
            setattr(obj, UID_ATTR, uid)
            items.append((id, obj))

            if do_action:
                # XXX NAAAAASTY Code. We need to create a special
                # workflow for Event.
                status = ev.getOne('status', default=None)
                actions = status_mapping.get(status, 'publish')
                for action in actions:
                    try:
                        wt.doActionFor(obj, action)
                    except WorkflowException:
                        # Try to revert to visible
                        for a in ('retract', 'show'):
                            try:
                                wt.doActionFor(obj, 'retract')
                            except WorkflowException:
                                pass
                        try:
                            wt.doActionFor(obj, action)
                        except WorkflowException:
                            pass
                        else:
                            break
                    else:
                        break
            obj.reindexObject(idxs=['Title',
                                    'start',
                                    'end'])
        return items


    def exportCalendar(self, events=None, REQUEST=None):
        wf = getToolByName(self, 'portal_workflow')
        ut = getToolByName(self, 'portal_url')
        if events is None:
            events = []
        dtstamp = datetime.utcnow().strftime("%Y%m%dT%H%M%SZ")
        uid_suffix = "%s" % ut()
        result = [
            "BEGIN:VCALENDAR",
            "PRODID:-//Runyaga.org/NONSGML Calendaring Tool//EN",
            "VERSION:1.0",
        ]

        uid_hash = None
        events.sort()
        for event in events:
            uid_hash = abs(hash(event))
            state = wf.getInfoFor(event, 'review_state')
            status = re_status_mapping.get(state, 'TENTATIVE')
            result += [
                "BEGIN:VEVENT",
                "UID:%s" % generateUID(event),
                "SUMMARY:%s" % ical_text(event.Title()),
                "DESCRIPTION:%s" % ical_text(event.Description())
                ]
            location = event.getLocation()
            if location:
                result += [
                    "LOCATION:%s" % ical_text(location)
                    ]
            if hasattr(aq_base(event), 'getAttendees'):
                for att in event.getAttendees():
                    name, email = parseaddr(att)
                    if not name:
                        name = email
                        email = ''
                    if email:
                        email = 'mailto:%s' % email
                    else:
                        email = 'invalid:nomail'
                    result += [
                        "ATTENDEE;CN=\"%s\":%s" % (
                        ical_text(name), email)
                        ]
            if event.duration == date.resolution:
                # All day event.
                result += [
                    "DTSTART;VALUE=DATE:%s" % event.start_date.strftime('%Y%m%d'),
                    "DTEND;VALUE=DATE:%s" % event.end_date.strftime('%Y%m%d'),
                    ]
            else:
                # Normal event
                result += [
                    "DTSTART:%s" % event.start_date.strftime('%Y%m%dT%H%M%S'),
                    "DTEND:%s" % event.end_date.strftime('%Y%m%dT%H%M%S'),
                    ]
            result += [
                "DTSTAMP:%s" % dtstamp,
                "STATUS:%s" % status,
                "END:VEVENT",
                ]
        if uid_hash is None:
            # There were no events.  iCalendar spec (RFC 2445) requires
            # VCALENDAR to have at least one subcomponent.  Let's create
            # a fake event.
            # NB Mozilla Calendar produces a 0-length file when publishing
            # empty calendars.  Sadly it does not then accept them
            # (http://bugzilla.mozilla.org/show_bug.cgi?id=229266).
            result += [
                "BEGIN:VEVENT",
                "UID:placeholder-%s" % uid_suffix,
                "SUMMARY:%s" % ical_text("Empty calendar"),
                "DTSTART:%s" % dtstamp[:-1],
                "DTSTAMP:%s" % dtstamp,
                "END:VEVENT",
            ]
        result.append("END:VCALENDAR")
        data = "\r\n".join(result)
        if REQUEST is not None:
            REQUEST.RESPONSE.setHeader('Content-Type',
                                       'text/calendar; charset=UTF-8')
            REQUEST.RESPONSE.write(data)
        return data

    security.declarePublic('getPublishedStates')
    def getPublishedStates(self):
        return self.getProperty('published_states')

    # catalog_getevents is copied verbatim from CMFCalendar
    # the only thing changed is the possiblility to set a published state
    security.declarePublic('catalog_getevents')
    def catalog_getevents(self, year, month):
        """ given a year and month return a list of days that have events """
        year=int(year)
        month=int(month)
        last_day=calendar.monthrange(year, month)[1]
        first_date=self.getBeginAndEndTimes(1, month, year)[0]
        last_date=self.getBeginAndEndTimes(last_day, month, year)[1]

        published = self.getPublishedStates()

        query=self.portal_catalog(portal_type=self.calendar_types,
                                  review_state=published,
                                  start={'query': last_date,
                                         'range': 'max'},
                                  end={'query': first_date,
                                       'range': 'min'},
                                  sort_on='start')

        # compile a list of the days that have events
        eventDays={}
        for daynumber in range(1, 32): # 1 to 31
            eventDays[daynumber] = {'eventslist':[], 'event':0, 'day':daynumber}
        includedevents = []
        for result in query:
            if result.getRID() in includedevents:
                break
            else:
                includedevents.append(result.getRID())
            event={}
            # we need to deal with events that end next month
            if  result.end.month() != month:  # doesn't work for events that last ~12 months - fix it if it's a problem, otherwise ignore
                eventEndDay = last_day
                event['end'] = None
            else:
                eventEndDay = result.end.day()
                event['end'] = result.end.Time()
            # and events that started last month
            if result.start.month() != month:  # same as above re: 12 month thing
                eventStartDay = 1
                event['start'] = None
            else:
                eventStartDay = result.start.day()
                event['start'] = result.start.Time()
            event['title'] = result.Title or result.id
            if eventStartDay != eventEndDay:
                allEventDays = range(eventStartDay, eventEndDay+1)
                eventDays[eventStartDay]['eventslist'].append({'end':None, 'start':result.start.Time(), 'title':result.Title})
                eventDays[eventStartDay]['event'] = 1
                for eventday in allEventDays[1:-1]:
                    eventDays[eventday]['eventslist'].append({'end':None, 'start':None, 'title':result.Title})
                    eventDays[eventday]['event'] = 1
                eventDays[eventEndDay]['eventslist'].append({'end':result.end.Time(), 'start':None, 'title':result.Title})
                eventDays[eventEndDay]['event'] = 1
            else:
                eventDays[eventStartDay]['eventslist'].append(event)
                eventDays[eventStartDay]['event'] = 1
            # This list is not uniqued and isn't sorted
            # uniquing and sorting only wastes time
            # and in this example we don't need to because
            # later we are going to do an 'if 2 in eventDays'
            # so the order is not important.
            # example:  [23, 28, 29, 30, 31, 23]
        return eventDays

InitializeClass(CalendarTool)
